# coding: utf-8
"""Work with volume annotations via the Supervisely API."""

import asyncio
import os
import re
from typing import Callable, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
from tqdm import tqdm

from supervisely._utils import batched
from supervisely.annotation.obj_class import ObjClass
from supervisely.api.entity_annotation.entity_annotation_api import EntityAnnotationAPI
from supervisely.api.module_api import ApiField
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.geometry.mask_3d import Mask3D
from supervisely.io.fs import (
    change_directory_at_index,
    dir_exists,
    file_exists,
    get_file_name,
    list_files,
    silent_remove,
)
from supervisely.io.json import load_json_file
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.volume import stl_converter
from supervisely.volume_annotation.volume_annotation import VolumeAnnotation
from supervisely.volume_annotation.volume_figure import VolumeFigure
from supervisely.volume_annotation.volume_object import VolumeObject
from supervisely.volume_annotation.volume_object_collection import (
    VolumeObjectCollection,
)


class VolumeAnnotationAPI(EntityAnnotationAPI):
    """API for working with volume annotations."""

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                api = sly.Api.from_env()
                ann_info = api.volume.annotation.download(volume_id)
        """
        super().__init__(api)

    _method_download_bulk = "volumes.annotations.bulk.info"
    _entity_ids_str = ApiField.VOLUME_IDS

    def download(self, volume_id: int):
        """
        Download information about VolumeAnnotation by volume ID from API.
        :param volume_id: Volume ID in Supervisely.
        :type volume_id: int
        :returns: Information about VolumeAnnotation in json format
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                volume_id = 19581134
                ann_info = api.volume.annotation.download(volume_id)
                print(ann_info)
                # Output:
                # {
                #     'createdAt': '2023-03-29T12:30:37.078Z',
                #     'datasetId': 61803,
                #     'description': '',
                #     'objects': [],
                #     'planes': [],
                #     'spatialFigures': [],
                #     'tags': [{'createdAt': '2023-04-03T13:21:53.368Z',
                #             'id': 12259702,
                #             'labelerLogin': 'almaz',
                #             'name': 'info',
                #             'tagId': 385328,
                #             'updatedAt': '2023-04-03T13:21:53.368Z',
                #             'value': 'age 31'}],
                #     'updatedAt': '2023-03-29T12:30:37.078Z',
                #     'volumeId': 19581134,
                #     'volumeMeta': {
                #             'ACS': 'RAS',
                #             'IJK2WorldMatrix': [0.7617, 0, 0,
                #                                 -194.2384, 0, 0.76171,
                #                                 0, -217.5384, 0,
                #                                 0, 2.5, -347.75,
                #                                 0, 0, 0, 1],
                #             'channelsCount': 1,
                #             'dimensionsIJK': {'x': 512, 'y': 512, 'z': 139},
                #             'intensity': {'max': 3071, 'min': -3024},
                #             'rescaleIntercept': 0,
                #             'rescaleSlope': 1,
                #             'windowCenter': 23.5,
                #             'windowWidth': 6095
                # },
                #     'volumeName': 'CTChest.nrrd'
                # }
        """

        volume_info = self._api.volume.get_info_by_id(volume_id)
        return self._download(volume_info.dataset_id, volume_id)

    def append(
        self, volume_id: int, ann: VolumeAnnotation, key_id_map: KeyIdMap = None, volume_info=None
    ):
        """
        Loads VolumeAnnotation to a given volume ID in the API.

        :param volume_id: Volume ID in Supervisely.
        :type volume_id: int
        :param ann: VolumeAnnotation object.
        :type ann: :class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation`
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                volume_id = 19581134
                api.volume.annotation.append(volume_id, volume_ann)
        """
        if ann.spatial_figures:
            figures = ann.figures + ann.spatial_figures
        else:
            figures = ann.figures

        if volume_info is None:
            volume_info = self._api.volume.get_info_by_id(volume_id)
        self._append(
            self._api.volume.tag,
            self._api.volume.object,
            self._api.volume.figure,
            volume_info.project_id,
            volume_info.dataset_id,
            volume_id,
            ann.tags,
            ann.objects,
            figures,
            key_id_map,
        )

    def upload_paths(
        self,
        volume_ids: List[int],
        ann_paths: List[str],
        project_meta: ProjectMeta,
        interpolation_dirs: Optional[List[str]] = None,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        mask_dirs: Optional[List[str]] = None,
    ) -> None:
        """
        Loads VolumeAnnotations from a given paths to a given volumes IDs in the API. Volumes IDs must be from one dataset.

        :param volume_ids: Volumes IDs in Supervisely.
        :type volume_ids: List[int]
        :param ann_paths: Paths to annotation files
        :type ann_paths: List[str]
        :param project_meta: Input ProjectMeta for VolumeAnnotations
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        :param interpolation_dirs: Paths to dirs with interpolation STL files
        :type interpolation_dirs: List[str], optional
        :param progress_cb: Function for tracking download progress
        :type progress_cb: tqdm or callable, optional
        :param mask_dirs: Paths to dirs with Mask3D geometries
        :type mask_dirs: List[str], optional
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                volume_ids = [121236918, 121236919]
                ann_pathes = ['/home/admin/work/supervisely/example/ann1.json', '/home/admin/work/supervisely/example/ann2.json']
                api.volume.annotation.upload_paths(volume_ids, ann_pathes, meta)
        """

        # use in updating project metadata
        project_id = self._api.volume.get_info_by_id(volume_ids[0]).project_id

        if interpolation_dirs is None:
            interpolation_dirs = [None] * len(ann_paths)

        if mask_dirs is None:
            mask_dirs = [None] * len(ann_paths)

        key_id_map = KeyIdMap()
        for volume_id, ann_path, interpolation_dir, mask_dir in zip(
            volume_ids, ann_paths, interpolation_dirs, mask_dirs
        ):
            ann_json = load_json_file(ann_path)
            ann = VolumeAnnotation.from_json(ann_json, project_meta)

            geometries_dict = {}
            stl_paths = []
            nrrd_paths = []
            keep_nrrd_paths = []

            if interpolation_dir is not None and dir_exists(interpolation_dir):
                (
                    stl_paths,
                    nrrd_paths,
                    keep_nrrd_paths,
                ) = self._prepare_convertation_paths(interpolation_dir, ann)

                if len(stl_paths) != 0:
                    stl_converter.to_nrrd(stl_paths, nrrd_paths)
                    ann, project_meta = self._update_on_transfer(
                        "upload", ann, project_meta, nrrd_paths, project_id
                    )

            # list all Mask3D geometries
            if mask_dir is not None and dir_exists(mask_dir):
                mask_paths = list_files(mask_dir, valid_extensions=[".nrrd"])
                geometries_dict.update(Mask3D._bytes_from_nrrd_batch(mask_paths))
                # it is not recommended to change the original composition of the project directory files
                # for this purpose delete the files created during conversion
                for nrrd_path in nrrd_paths:
                    if nrrd_path not in keep_nrrd_paths:
                        silent_remove(nrrd_path)

            # add geometries into spatial figure objects
            for sf in ann.spatial_figures:
                try:
                    geometry_bytes = geometries_dict[sf.key().hex]
                    mask3d = Mask3D.from_bytes(geometry_bytes)
                    sf._set_3d_geometry(mask3d)
                except (KeyError, TypeError) as e:
                    if isinstance(e, TypeError):
                        logger.warning(
                            f"Skipping spatial figure for class '{sf.volume_object.obj_class.name}': {str(e)}"
                        )
                    # skip figures that doesn't need to update geometry
                    # for example for old geometries that are stored in JSON (KeyError)
                    continue

            self.append(volume_id, ann, key_id_map)

            if progress_cb is not None:
                progress_cb(1)

    def _update_on_transfer(
        self,
        transfer_type: Literal["download", "upload"],
        ann: VolumeAnnotation,
        project_meta: ProjectMeta,
        nrrd_paths: List[str],
        project_id: Optional[int] = None,
    ) -> Tuple[VolumeAnnotation, ProjectMeta]:
        """
        Create new ObjClass and VolumeFigure annotations for converted STL.
        Replace ClosedMeshSurface spatial figures with Mask3D.
        Update the ann, project_meta, and key_id_map.

        :param transfer_type: Defines the process during which the update will be performed ("download" or "upload").
        :type transfer_type: Literal["download", "upload"]
        :param ann: VolumeAnnotation object to update.
        :type ann: :class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation`
        :param project_meta: ProjectMeta object.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        :param nrrd_paths: Paths to the converted NRRD files from STL.
        :type nrrd_paths: List[str]
        :param project_id: The Project ID to update metadata on upload (optional).
        :type project_id: int, optional
        :returns: A tuple containing the updated ann and project_meta objects.
        :rtype: Tuple[:class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation`, :class:`~supervisely.project.project_meta.ProjectMeta`]
        """

        for nrrd_path in nrrd_paths:
            object_key = None
            custom_data = None

            # searching connection between interpolation and spatial figure in annotations and set its object_key
            for sf in ann.spatial_figures:
                if sf.key().hex == get_file_name(nrrd_path):
                    object_key = sf.parent_object.key()
                    custom_data = sf.custom_data
                    break

            if object_key:
                for obj in ann.objects:
                    if obj.key() == object_key:
                        class_title = obj.obj_class.name
                        break

            else:
                raise Exception(
                    f"Can't find volume object for Mask3D from unknown file '{nrrd_path}'. Please check the project structure."
                )

            class_created = False
            new_obj_class = project_meta.get_obj_class(f"{class_title}_mask_3d")

            if new_obj_class is None:
                new_obj_class = ObjClass(f"{class_title}_mask_3d", Mask3D)
                project_meta = project_meta.add_obj_class(new_obj_class)
                class_created = True

            geometry = Mask3D(np.zeros((3, 3, 3), dtype=np.bool_))

            if transfer_type == "download":
                new_object = VolumeObject(new_obj_class, mask_3d=geometry, custom_data=custom_data)
            elif transfer_type == "upload":
                if class_created:
                    self._api.project.update_meta(project_id, project_meta)
                new_object = VolumeObject(new_obj_class)
                new_object.figure = VolumeFigure(new_object, geometry, key=sf.key(), custom_data=custom_data)

            # add new Volume object to VolumeAnnotation with spatial figure
            ann = ann.add_objects([new_object])

            if transfer_type == "download":
                # as a sf.key() changes, we need to rename both files
                os.rename(
                    nrrd_path, f"{os.path.dirname(nrrd_path)}/{new_object.figure.key().hex}.nrrd"
                )
                stl_path = re.sub(r"\.[^.]+$", ".stl", nrrd_path)
                stl_path = change_directory_at_index(stl_path, "interpolation", -3)
                os.rename(
                    stl_path, f"{os.path.dirname(stl_path)}/{new_object.figure.key().hex}.stl"
                )
            # remove STL spatial figure from VolumeAnnotation
            if sf:
                ann.spatial_figures.remove(sf)

        return ann, project_meta

    def append_objects(
        self,
        volume_id: int,
        objects: Union[List[VolumeObject], VolumeObjectCollection],
        key_id_map: Optional[KeyIdMap] = None,
    ) -> None:
        """
        Add new VolumeObjects to a volume annotation in Supervisely project.

        :param volume_id: The ID of the volume.
        :type volume_id: int
        :param objects: New volume objects.
        :type objects: List[:class:`~supervisely.volume_annotation.volume_object.VolumeObject`] or :class:`~supervisely.volume_annotation.volume_object_collection.VolumeObjectCollection`
        :param key_id_map: KeyIdMap object (optional).
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: None
        :rtype: NoneType

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                volume_id = 151344
                volume_info = api.volume.get_info_by_id(volume_id)
                mask_3d_path = "data/mask/lung.nrrd"
                lung_obj_class = sly.ObjClass("lung", sly.Mask3D)
                lung = sly.VolumeObject(lung_obj_class, mask_3d=mask_3d_path)
                objects = sly.VolumeObjectCollection([lung])
                api.volume.annotation.append_objects(volume_info.id, objects)
        """

        sf_figures = []
        for volume_object in objects:
            if volume_object.obj_class.geometry_type in (Mask3D, AnyGeometry):
                if isinstance(volume_object.figure.geometry, Mask3D):
                    sf_figures.append(volume_object.figure)

        volume_meta = self._api.volume.get_info_by_id(volume_id).meta
        ann = VolumeAnnotation(volume_meta, objects, spatial_figures=sf_figures)
        self.append(volume_id, ann, key_id_map)

    @staticmethod
    def _prepare_convertation_paths(
        interpolation_dir: str, ann: VolumeAnnotation
    ) -> Tuple[List, List, List]:
        """
        Check dir and create paths for NRRD files if STL files need to be converted.

        :param interpolation_dir: Path to dir with interpolation STL files
        :type interpolation_dir: str
        :param ann: VolumeAnnotation object
        :type ann: :class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation`
        :returns: Paths to STL and NRRD files used in the conversion process
        :rtype: Tuple[List, List, List]
        """
        stl_paths_in = list_files(interpolation_dir, valid_extensions=[".stl"])
        nrrd_paths = []
        stl_paths = []
        keep_nrrd_paths = []  # to keep original composition of the project
        for stl_path in stl_paths_in:
            # check if this is really STL file
            with open(stl_path, "rb") as file:
                header = file.read(84)
                if b"solid" in header:
                    stl_paths.append(stl_path)
                else:
                    continue
            nrrd_path = re.sub(r"\.[^.]+$", ".nrrd", stl_path)
            nrrd_path = change_directory_at_index(nrrd_path, "mask", -3)
            nrrd_paths.append(nrrd_path)
            if file_exists(nrrd_path):
                keep_nrrd_paths.append(nrrd_path)
                # to prevent duplication if STL is already converted to NRRD on export
                for sf in ann.spatial_figures:
                    if sf.key().hex == get_file_name(nrrd_path) and isinstance(sf.geometry, Mask3D):
                        stl_paths.remove(stl_path)
                        nrrd_paths.remove(nrrd_path)
                        keep_nrrd_paths.remove(nrrd_path)
        return stl_paths, nrrd_paths, keep_nrrd_paths

    async def download_async(
        self,
        volume_id: int,
        semaphore: Optional[asyncio.Semaphore] = None,
        integer_coords: bool = True,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> Dict:
        """
        Download information about VolumeAnnotation by volume ID from API asynchronously.

        :param volume_id: Volume ID in Supervisely.
        :type volume_id: int
        :param semaphore: Semaphore to limit the number of parallel downloads.
        :type semaphore: asyncio.Semaphore, optional
        :param integer_coords: If True, returns coordinates as integers for objects. If False, returns as floats.
        :type integer_coords: bool, optional
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :returns: Information about VolumeAnnotation in json format
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                volume_id = 198702499
                loop = sly.utils.get_or_create_event_loop()
                ann_info = loop.run_until_complete(api.volume.annotation.download_async(volume_id))
        """
        return await self.download_bulk_async(
            volume_ids=[volume_id],
            semaphore=semaphore,
            integer_coords=integer_coords,
            progress_cb=progress_cb,
        )

    async def download_bulk_async(
        self,
        volume_ids: List[int],
        semaphore: Optional[asyncio.Semaphore] = None,
        integer_coords: bool = True,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> List[Dict]:
        """
        Download information about VolumeAnnotation in bulk by volume IDs from API asynchronously.

        :param volume_ids: List of Volume IDs in Supervisely. All volumes must be from the same dataset.
        :type volume_ids: int
        :param semaphore: Semaphore to limit the number of parallel downloads.
        :type semaphore: asyncio.Semaphore, optional
        :param integer_coords: If True, returns coordinates as integers for objects. If False, returns as floats.
        :type integer_coords: bool, optional
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :returns: Information about VolumeAnnotations in json format
        :rtype: list

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                volume_ids = [198702499, 198702500, 198702501]
                loop = sly.utils.get_or_create_event_loop()
                ann_infos = loop.run_until_complete(api.volume.annotation.download_bulk_async(volume_ids))
        """
        if semaphore is None:
            semaphore = self._api.get_default_semaphore()

        async def fetch_with_semaphore(batch):
            async with semaphore:
                json_data = {
                    self._entity_ids_str: batch,
                    ApiField.INTEGER_COORDS: integer_coords,
                }
                response = await self._api.post_async(
                    self._method_download_bulk,
                    json=json_data,
                )
                if progress_cb is not None:
                    progress_cb(len(batch))
                return response.json()

        tasks = [fetch_with_semaphore(batch) for batch in batched(volume_ids)]
        responses = await asyncio.gather(*tasks)
        json_response = [item for response in responses for item in response]
        return json_response



# coding: utf-8

from supervisely.video_annotation.video_tag_collection import VideoTagCollection
from supervisely.volume_annotation.volume_tag import VolumeTag


class VolumeTagCollection(VideoTagCollection):
    """
    Collection with :class:`~supervisely.volume_annotation.volume_tag.VolumeTag` instances. :class:`~supervisely.volume_annotation.volume_tag_collection.VolumeTagCollection` object is immutable.

    :Usage Example:

        .. code-block:: python

            import supervisely as sly
            from supervisely.volume_annotation.volume_tag import VolumeTag
            from supervisely.volume_annotation.volume_tag_collection import VolumeTagCollection

            # Create two VolumeTags for collection
            meta_brain = sly.TagMeta('brain_tag', sly.TagValueType.ANY_STRING)
            brain_tag = VolumeTag(meta_brain, value='brain')
            meta_heart = sly.TagMeta('heart_tag', sly.TagValueType.ANY_STRING)
            heart_tag = VolumeTag(meta_heart, value='heart')

            # Create VolumeTagCollection
            tags = VolumeTagCollection([brain_tag, heart_tag])
            tags_json = tags.to_json()
            print(tags_json)
            # Output:
            # [
            #     {
            #         "key": "9fbbc3f888594a538243445fe25242ec",
            #         "name": "brain_tag",
            #         "value": "brain"
            #     },
            #     {
            #         "key": "804c8124d46c4da89b54c6132acf06a0",
            #         "name": "heart_tag",
            #         "value": "heart"
            #     }
            # ]


            # Add item to VolumeTagCollection
            meta_lang = sly.TagMeta('lang_tag', sly.TagValueType.NONE)
            lang_tag = VolumeTag(meta_lang)
            # Remember that VolumeTagCollection is immutable, and we need to assign new instance of VolumeTagCollection to a new variable
            new_tags = tags.add(lang_tag)
            new_tags_json = new_tags.to_json()
            print(new_tags_json)
            # Output: [
            #     {
            #     {
            #         "key": "9fbbc3f888594a538243445fe25242ec",
            #         "name": "brain_tag",
            #         "value": "brain"
            #     },
            #     {
            #         "key": "804c8124d46c4da89b54c6132acf06a0",
            #         "name": "heart_tag",
            #         "value": "heart"
            #     },
            #     {
            #         "key": "7188242d2ddb4d2783c588cc2eca5ff8",
            #         "name": "lang_tag"
            #     }
            # ]

            # You can also add multiple items to collection
            meta_leg = sly.TagMeta('leg_tag', sly.TagValueType.NONE)
            leg_tag = VolumeTag(meta_leg)
            meta_arm = sly.TagMeta('arm_tag', sly.TagValueType.ANY_NUMBER)
            arm_tag = VolumeTag(meta_arm, value=777)
            new_tags = tags.add_items([leg_tag, arm_tag])
            new_tags_json = new_tags.to_json()
            print(new_tags_json)
            # Output: [
            #     {
            #     {
            #         "key": "9fbbc3f888594a538243445fe25242ec",
            #         "name": "brain_tag",
            #         "value": "brain"
            #     },
            #     {
            #         "key": "804c8124d46c4da89b54c6132acf06a0",
            #         "name": "heart_tag",
            #         "value": "heart"
            #     },
            #     {
            #         "key": "7188242d2ddb4d2783c588cc2eca5ff8",
            #         "name": "lang_tag"
            #     },
            #     {
            #         "key": "111c8124d46c4da89b5lgk132acf06a0",
            #         "name": "leg_tag",
            #     },
            #     {
            #         "key": "28dc8124d46c4da89b54c6132acf06a0",
            #         "name": "arm_tag",
            #         "value": "777"
            #     },
            # ]

            # Find intersection of given list of VolumeTag instances with collection items
            intersect_tags = tags.intersection([leg_tag])
            intersect_tags_json = intersect_tags.to_json()
            print(intersect_tags_json)
            # Output: [
            #     {
            #         "key": "111c8124d46c4da89b5lgk132acf06a0",
            #         "name": "leg_tag",
            #     }
            # ]

            # Find difference between collection and given list of VolumeTag
            diff_tags = tags.difference([leg_tag, arm_tag, lang_tag])
            diff_tags_json = diff_tags.to_json()
            print(diff_tags_json)
            # Output:
            # [
            #     {
            #         "key": "9fbbc3f888594a538243445fe25242ec",
            #         "name": "brain_tag",
            #         "value": "brain"
            #     },
            #     {
            #         "key": "804c8124d46c4da89b54c6132acf06a0",
            #         "name": "heart_tag",
            #         "value": "heart"
            #     }
            # ]

            # Merge collection and given list of VolumeTagCollection
            meta_leg = sly.TagMeta('leg_tag', sly.TagValueType.NONE)
            leg_tag = VolumeTag(meta_leg)
            meta_arm = sly.TagMeta('arm_tag', sly.TagValueType.ANY_NUMBER)
            arm_tag = VolumeTag(meta_arm, value=777)
            over_tags = VolumeTagCollection([leg_tag, arm_tag])
            # Merge
            merge_tags = tags.merge(over_tags)
            merge_tags_json = merge_tags.to_json()
            print(merge_tags_json)
            # Output: [
            # [
            #     {
            #         "key": "9fbbc3f888594a538243445fe25242ec",
            #         "name": "brain_tag",
            #         "value": "brain"
            #     },
            #     {
            #         "key": "804c8124d46c4da89b54c6132acf06a0",
            #         "name": "heart_tag",
            #         "value": "heart"
            #     },
            #     {
            #         "key": "111c8124d46c4da89b5lgk132acf06a0",
            #         "name": "leg_tag",
            #     },
            #     {
            #         "key": "28dc8124d46c4da89b54c6132acf06a0",
            #         "name": "arm_tag",
            #         "value": "777"
            #     }
            # ]
    """

    item_type = VolumeTag



from typing import Literal
from supervisely.app.widgets import Widget, Progress
from supervisely.app.content import DataJson


class CircleProgress(Widget):
    """Circular progress indicator widget showing Progress status (success/exception/none)."""

    def __init__(self, progress: Progress, widget_id=None):
        """
        :param progress: Progress object to display status for.
        :type progress: :class:`~supervisely.app.widgets.progress.Progress`
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self.progress = progress
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "status": None,
        }

    def get_json_state(self):
        return {}

    def set_status(self, status: Literal["success", "exception", "none"]):
        if status == "none":
            status = None
        DataJson()[self.widget_id]["status"] = status
        DataJson().send_changes()



# coding: utf-8
"""Get a list of plugins for a Supervisely team."""

from __future__ import annotations

from typing import Dict, List, NamedTuple, Optional

from supervisely.api.module_api import ApiField, ModuleApi


class PluginApi(ModuleApi):
    """API for working with plugins."""

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                api = sly.Api.from_env()
                plugin_info = api.plugin.get_list(team_id)
        """
        super().__init__(api)

    @staticmethod
    def info_sequence():
        """
        NamedTuple PluginInfo information about Plugin.

        :Usage Example:

            .. code-block:: python

                PluginInfo(
                    id=3,
                    name='DTL',
                    description='Allows to combine datasets, to make class mapping, filter objects and images, apply auto augmentations and so on ...',
                    type='dtl',
                    default_version='latest',
                    docker_image='docker.deepsystems.io/supervisely/five/dtl',
                    readme='# Data Transformation Language (DTL)...',
                    configs=[],
                    versions=['lately', 'docs', ...],
                    created_at='2020-03-30T09:17:36.000Z',
                    updated_at='2020-04-23T06:26:29.000Z'
                )
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.DESCRIPTION,
            ApiField.TYPE,
            ApiField.DEFAULT_VERSION,
            ApiField.DOCKER_IMAGE,
            ApiField.README,
            ApiField.CONFIGS,
            ApiField.VERSIONS,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **PluginInfo**.
        """
        return "PluginInfo"

    def get_list(
        self, team_id: int, filters: Optional[List[Dict[str, str]]] = None
    ) -> List[NamedTuple]:
        """
        Get list of plugins in the Team.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param filters: List of params to sort output Plugins.
        :type filters: List[dict], optional
        :returns: List of Plugins with information.
        :rtype: :class:`List[NamedTuple]`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                team_id = 8
                plugin_info = api.plugin.get_list(team_id)

                plugin_list_filter = api.plugin.get_list(team_id, filters=[{'field': 'name', 'operator': '=', 'value': 'Images'}])
        """
        return self.get_list_all_pages(
            "plugins.list", {ApiField.TEAM_ID: team_id, ApiField.FILTER: filters or []}
        )

    def get_info_by_id(self, team_id: int, plugin_id: int) -> NamedTuple:
        """
        Get Plugin information by ID.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param plugin_id: Plugin ID in Supervisely.
        :type plugin_id: int
        :returns: Information about Plugin.
        :rtype: :class:`NamedTuple`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                plugin_info = api.plugin.get_info_by_id(8, 3)
                print(plugin_info)
                # Output: PluginInfo(id=3,
                #                    name='DTL',
                #                    description='Allows to combine datasets, to make class mapping, filter objects and images, apply auto augmentations and so on ...',
                #                    type='dtl',
                #                    default_version='latest',
                #                    docker_image='docker.deepsystems.io/supervisely/five/dtl',
                #                    readme='# Data Transformation Language (DTL)...',
                #                    configs=[],
                #                    versions=['lately', 'docs', ...],
                #                    created_at='2020-03-30T09:17:36.000Z',
                #                    updated_at='2020-04-23T06:26:29.000Z')
        """
        filters = [{"field": ApiField.ID, "operator": "=", "value": plugin_id}]
        return self._get_info_by_filters(team_id, filters)



from typing import List, Union, Dict, Tuple
from pathlib import Path
from collections import defaultdict
import numpy as np

import supervisely as sly
from supervisely.nn.model.prediction import Prediction
from supervisely.annotation.label import LabelingStatus
from supervisely import VideoAnnotation
from supervisely import logger


def predictions_to_video_annotation(
    predictions: List[Prediction],
) -> VideoAnnotation:
    """
    Convert list of Prediction objects to VideoAnnotation.

    :param predictions: List of Prediction objects, one per frame
    :type predictions: List[Prediction]
    :returns: VideoAnnotation object with tracked objects
    :rtype: VideoAnnotation
    """

    if not predictions:
        raise ValueError("Empty predictions list provided")

    frame_shape = predictions[0].annotation.img_size
    img_h, img_w = frame_shape
    video_objects = {}  
    frames = []

    for pred in predictions:
        frame_figures = []
        frame_idx = pred.frame_index

        # Get data using public properties
        boxes = pred.boxes          # Public property - np.array (N, 4) in tlbr format
        classes = pred.classes      # Public property - list of class names
        track_ids = pred.track_ids  # Public property - can be None

        # Skip frame if no detections
        if len(boxes) == 0:
            frames.append(sly.Frame(frame_idx, []))
            continue

        for bbox, class_name, track_id in zip(boxes, classes, track_ids):
            # Clip bbox to image boundaries
            # Note: pred.boxes returns tlbr format (top, left, bottom, right)
            top, left, bottom, right = bbox
            dims = np.array([img_h, img_w, img_h, img_w]) - 1
            top, left, bottom, right = np.clip([top, left, bottom, right], 0, dims)

            # Convert to integer coordinates
            top, left, bottom, right = int(top), int(left), int(bottom), int(right)

            # Get or create VideoObject
            if track_id not in video_objects:
                # Find obj_class from prediction annotation
                obj_class = None
                for label in pred.annotation.labels:
                    if label.obj_class.name == class_name:
                        obj_class = label.obj_class
                        break

                if obj_class is None:
                    # Create obj_class if not found (fallback)
                    obj_class = sly.ObjClass(class_name, sly.Rectangle)

                video_objects[track_id] = sly.VideoObject(obj_class)

            video_object = video_objects[track_id]
            rect = sly.Rectangle(top=top, left=left, bottom=bottom, right=right)
            frame_figures.append(sly.VideoFigure(video_object, rect, frame_idx, track_id=str(track_id), status=LabelingStatus.AUTO))

        frames.append(sly.Frame(frame_idx, frame_figures))

    objects = list(video_objects.values()) 
    return VideoAnnotation(
        img_size=frame_shape,
        frames_count=len(predictions),
        objects=sly.VideoObjectCollection(objects),
        frames=sly.FrameCollection(frames)
    )

def video_annotation_to_mot(
    annotation: VideoAnnotation,
    output_path: Union[str, Path] = None,
    class_to_id_mapping: Dict[str, int] = None
) -> Union[str, List[str]]:
    """
    Convert Supervisely VideoAnnotation to MOT format.
    MOT format: frame_id,track_id,left,top,width,height,confidence,class_id,visibility

    :param annotation: VideoAnnotation object to convert
    :type annotation: VideoAnnotation
    :param output_path: Path to save the MOT file
    :type output_path: Union[str, Path]
    :param class_to_id_mapping: Mapping of class names to IDs
    :type class_to_id_mapping: Dict[str, int]
    :returns: Path to saved MOT file or list of MOT lines
    :rtype: Union[str, List[str]]
    """
    mot_lines = []

    # Create default class mapping if not provided
    if class_to_id_mapping is None:
        unique_classes = set()
        for frame in annotation.frames:
            for figure in frame.figures:
                unique_classes.add(figure.video_object.obj_class.name)
        class_to_id_mapping = {cls_name: idx + 1 for idx, cls_name in enumerate(sorted(unique_classes))}

    # Extract tracks
    for frame in annotation.frames:
        frame_id = frame.index + 1  # MOT uses 1-based frame indexing

        for figure in frame.figures:
            # Get track ID from VideoFigure.track_id (official API)
            if figure.track_id is not None:
                track_id = int(figure.track_id)
            else:
                track_id = figure.video_object.key().int

            # Get bounding box
            if isinstance(figure.geometry, sly.Rectangle):
                bbox = figure.geometry
            else:
                bbox = figure.geometry.to_bbox()

            left = bbox.left
            top = bbox.top
            width = bbox.width
            height = bbox.height

            # Get class ID
            class_name = figure.video_object.obj_class.name
            class_id = class_to_id_mapping.get(class_name, 1)

            # Get confidence (default)
            confidence = 1.0

            # Visibility (assume visible)
            visibility = 1

            # Create MOT line
            mot_line = f"{frame_id},{track_id},{left:.2f},{top:.2f},{width:.2f},{height:.2f},{confidence:.3f},{class_id},{visibility}"
            mot_lines.append(mot_line)

    # Save to file if path provided
    if output_path:
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)

        with open(output_path, 'w') as f:
            for line in mot_lines:
                f.write(line + '\n')

        logger.info(f"Saved MOT format to: {output_path} ({len(mot_lines)} detections)")
        return str(output_path)

    return mot_lines

def mot_to_video_annotation(
    mot_file_path: Union[str, Path],
    img_size: Tuple[int, int] = (1080, 1920),
    class_mapping: Dict[int, str] = None,
    default_class_name: str = "person"
) -> VideoAnnotation:
    """
    Convert MOT format tracking data to Supervisely VideoAnnotation.
    MOT format: frame_id,track_id,left,top,width,height,confidence,class_id,visibility

    :param mot_file_path: Path to the MOT file
    :type mot_file_path: Union[str, Path]
    :param img_size: Image size (height, width)
    :type img_size: Tuple[int, int]
    :param class_mapping: Mapping of class IDs to names
    :type class_mapping: Dict[int, str]
    :param default_class_name: Default class name if not found
    :type default_class_name: str
    :returns: VideoAnnotation object with tracked objects
    :rtype: :class:`~supervisely.video_annotation.video_annotation.VideoAnnotation`
    """
    mot_file_path = Path(mot_file_path)

    if not mot_file_path.exists():
        raise FileNotFoundError(f"MOT file not found: {mot_file_path}")

    logger.info(f"Loading MOT data from: {mot_file_path}")
    logger.info(f"Image size: {img_size} (height, width)")

    # Default class mapping
    if class_mapping is None:
        class_mapping = {1: default_class_name}

    # Parse MOT file
    video_objects = {}  # track_id -> VideoObject
    frames_data = defaultdict(list)  # frame_idx -> list of figures
    max_frame_idx = 0
    img_h, img_w = img_size

    with open(mot_file_path, 'r') as f:
        for line_num, line in enumerate(f, 1):
            line = line.strip()
            if not line or line.startswith('#'):
                continue

            try:
                parts = line.split(',')
                if len(parts) < 6:  # Minimum required fields
                    continue

                frame_id = int(parts[0])
                track_id = int(parts[1])
                left = float(parts[2])
                top = float(parts[3])
                width = float(parts[4])
                height = float(parts[5])

                # Optional fields
                confidence = float(parts[6]) if len(parts) > 6 and parts[6] != '-1' else 1.0
                class_id = int(parts[7]) if len(parts) > 7 and parts[7] != '-1' else 1
                visibility = float(parts[8]) if len(parts) > 8 and parts[8] != '-1' else 1.0

                frame_idx = frame_id - 1  # Convert to 0-based indexing
                max_frame_idx = max(max_frame_idx, frame_idx)

                # Skip low confidence detections
                if confidence < 0.1:
                    continue

                # Calculate coordinates with safer clipping
                right = left + width
                bottom = top + height

                # Clip to image boundaries
                left = max(0, int(left))
                top = max(0, int(top))
                right = min(int(right), img_w - 1)
                bottom = min(int(bottom), img_h - 1)

                # Skip invalid boxes
                if right <= left or bottom <= top:
                    continue

                # Get class name
                class_name = class_mapping.get(class_id, default_class_name)

                # Create VideoObject if not exists
                if track_id not in video_objects:
                    obj_class = sly.ObjClass(class_name, sly.Rectangle)
                    video_objects[track_id] = sly.VideoObject(obj_class)

                video_object = video_objects[track_id]

                # Create rectangle and figure with track_id
                rect = sly.Rectangle(top=top, left=left, bottom=bottom, right=right)
                figure = sly.VideoFigure(video_object, rect, frame_idx, track_id=str(track_id))

                frames_data[frame_idx].append(figure)

            except (ValueError, IndexError) as e:
                logger.warning(f"Skipped invalid MOT line {line_num}: {line} - {e}")
                continue

    # Create frames
    frames = []
    if frames_data:
        frames_count = max(frames_data.keys()) + 1

        for frame_idx in range(frames_count):
            figures = frames_data.get(frame_idx, [])
            frames.append(sly.Frame(frame_idx, figures))
    else:
        frames_count = 1
        frames = [sly.Frame(0, [])]

    # Create VideoAnnotation
    objects = list(video_objects.values())

    annotation = VideoAnnotation(
        img_size=img_size,
        frames_count=frames_count,
        objects=sly.VideoObjectCollection(objects),
        frames=sly.FrameCollection(frames)
    )

    logger.info(f"Created VideoAnnotation with {len(objects)} tracks and {frames_count} frames")

    return annotation



from typing import Any, List, Optional
from supervisely.app.widgets import Widget


class OptionComponent:
    """Base option component configuration used by NodesFlow node settings."""

    def __init__(
        self,
        component_name: str,
        default_value: Optional[Any] = None,
        sidebar_component: Optional[str] = None,
        options: Optional[dict] = {},
    ):
        """:param component_name: Name of the component type.
        :type component_name: str
        :param default_value: Default value for the option.
        :type default_value: Any, optional
        :param sidebar_component: Optional sidebar component name.
        :type sidebar_component: str, optional
        :param options: Additional options dict.
        :type options: dict, optional
        """
        self.component_name = component_name
        self.default_value = default_value
        self.sidebar_component = sidebar_component
        self.options = options

    def to_json(self):
        return {
            "component": self.component_name,
            "value": self.default_value,
            "sidebarComponent": self.sidebar_component,
            "options": self.options,
        }


class HtmlOptionComponent(OptionComponent):
    """Option component rendered from arbitrary HTML (optionally with a sidebar component)."""

    def __init__(
        self,
        html: str,
        sidebar_component: Optional[OptionComponent] = None,
        sidebar_width: Optional[int] = None,
    ):
        """:param html: HTML string for the option.
        :type html: str
        :param sidebar_component: Optional sidebar option component.
        :type sidebar_component: OptionComponent, optional
        :param sidebar_width: Width of sidebar in pixels.
        :type sidebar_width: int, optional
        """
        options = {}
        sidebar_component_name = None
        if sidebar_component is not None:
            sidebar_component_name = sidebar_component.component_name
            sidebar_template = sidebar_component.options["template"]
            options["sidebarTemplate"] = f"<div>{sidebar_template}</div>"
            options["sidebarWidth"] = sidebar_width
        options["template"] = f"<div>{html}</div>"
        super().__init__(
            component_name="SlyFlowOptionRenderer",
            options=options,
            sidebar_component=sidebar_component_name
        )


class WidgetOptionComponent(HtmlOptionComponent):
    """Option component rendered from a Supervisely widget's HTML."""

    def __init__(self, widget: Widget, sidebar_component=None, sidebar_width: Optional[int] = None):
        """:param widget: Widget to render as option.
        :type widget: Widget
        :param sidebar_component: Optional sidebar component.
        :param sidebar_width: Sidebar width in pixels.
        :type sidebar_width: int, optional
        """
        super().__init__(
            widget.to_html(), sidebar_component=sidebar_component, sidebar_width=sidebar_width
        )


class ButtonOptionComponent(OptionComponent):
    """A button that opens the sidebar when clicked. The label of the button is determined by the name of the option."""

    def __init__(
        self,
        sidebar_component: Optional[OptionComponent] = None,
        sidebar_width: Optional[int] = None,
    ):
        """:param sidebar_component: Component to show in sidebar when clicked.
        :type sidebar_component: OptionComponent, optional
        :param sidebar_width: Sidebar width in pixels.
        :type sidebar_width: int, optional
        """
        sidebar_component_name = None
        options = {}
        if sidebar_component is not None:
            sidebar_component_name = sidebar_component.component_name
            options = sidebar_component.options
            options["sidebarWidth"] = sidebar_width
        super().__init__(
            component_name="ButtonOption", sidebar_component=sidebar_component_name, options=options
        )


class CheckboxOptionComponent(OptionComponent):
    """A checkbox for setting boolean values."""

    def __init__(self, default_value: bool = False):
        """:param default_value: Initial checked state.
        :type default_value: bool
        """
        super().__init__(component_name="CheckboxOption", default_value=default_value)


class InputOptionComponent(OptionComponent):
    """A simple text field. The option name will be displayed as placeholder."""

    def __init__(self, default_value: str = ""):
        """:param default_value: Initial text value.
        :type default_value: str
        """
        super().__init__(component_name="InputOption", default_value=default_value)


class IntegerOptionComponent(OptionComponent):
    """A numeric up/down field for integer values."""

    def __init__(
        self,
        min: Optional[int] = None,
        max: Optional[int] = None,
        default_value: Optional[int] = 0,
    ):
        """:param min: Minimum value.
        :type min: int, optional
        :param max: Maximum value.
        :type max: int, optional
        :param default_value: Initial value.
        :type default_value: int, optional
        """
        options = {}
        if min is not None:
            options["min"] = min
        if max is not None:
            options["max"] = max
        super().__init__(
            component_name="IntegerOption",
            default_value=default_value,
            options=options,
        )


class NumberOptionComponent(OptionComponent):
    """A numeric up/down field for numeric values."""

    def __init__(
        self,
        min: Optional[float] = None,
        max: Optional[float] = None,
        default_value: Optional[float] = 0,
    ):
        """:param min: Minimum value.
        :type min: float, optional
        :param max: Maximum value.
        :type max: float, optional
        :param default_value: Initial value.
        :type default_value: float, optional
        """
        options = {}
        if min is not None:
            options["min"] = min
        if max is not None:
            options["max"] = max
        super().__init__(
            component_name="NumberOption",
            default_value=default_value,
            options=options,
        )


class SelectOptionComponent(OptionComponent):
    """A dropdown select which allows the user to choose one of predefined values."""

    class Item:
        """Select option item (value + display label)."""

        def __init__(self, value: str, label: Optional[str] = None):
            """:param value: Option value.
            :type value: str
            :param label: Display label. Defaults to value.
            :type label: str, optional
            """
            self.value = value
            self.label = value if label is None else label

        def to_json(self):
            return {"text": self.label, "value": self.value}

    def __init__(self, items: List[Item], default_value: Optional[str] = None):
        """:param items: List of Item (value, label).
        :type items: List[Item]
        :param default_value: Initially selected value.
        :type default_value: str, optional
        """
        options = {"items": [item.to_json() for item in items]}
        super().__init__(
            component_name="SelectOption",
            default_value=default_value,
            options=options,
        )


class SliderOptionComponent(OptionComponent):
    """A slider for choosing a value in a range. The user can also click on the slider to set a specific value."""

    def __init__(
        self,
        min: float,
        max: float,
        default_value: Optional[float] = None,
    ):
        """:param min: Minimum value.
        :type min: float
        :param max: Maximum value.
        :type max: float
        :param default_value: Initial value. Defaults to min.
        :type default_value: float, optional
        """
        options = {
            "min": min,
            "max": max,
        }
        default_value = min if default_value is None else default_value
        super().__init__(
            component_name="SliderOption",
            default_value=default_value,
            options=options,
        )


class TextOptionComponent(OptionComponent):
    """Displays arbitrary strings"""

    def __init__(self, text: str):
        """:param text: Static text to display.
        :type text: str
        """
        super().__init__(component_name="TextOption", default_value=text)


class SidebarNodeInfoOptionComponent(OptionComponent):
    """Special option component used to render node info in the sidebar."""

    def __init__(self, sidebar_template: str, sidebar_width: Optional[int] = None):
        """:param sidebar_template: HTML template for sidebar content.
        :type sidebar_template: str
        :param sidebar_width: Sidebar width in pixels.
        :type sidebar_width: int, optional
        """
        options = {
            "sidebarTemplate": f"<div>{sidebar_template}</div>", 
            "sidebarWidth": sidebar_width
        }
        super().__init__(
            component_name="SlyFlowOptionRenderer",
            sidebar_component="SlyFlowOptionRenderer",
            options=options,
        )



import random
from pathlib import Path
from typing import List, Optional, Union

from supervisely.api.api import Api
from supervisely.app.widgets import SlyTqdm
from supervisely.imaging.color import get_predefined_colors, rgb2hex
from supervisely.io import env
from supervisely.io.fs import dir_empty, mkdir
from supervisely.io.json import load_json_file
from supervisely.nn.benchmark.comparison.detection_visualization.visualizer import (
    DetectionComparisonVisualizer,
)
from supervisely.nn.benchmark.comparison.semantic_segmentation.visualizer import (
    SemanticSegmentationComparisonVisualizer,
)
from supervisely.nn.benchmark.cv_tasks import CVTask
from supervisely.nn.benchmark.object_detection.evaluator import (
    ObjectDetectionEvalResult,
)
from supervisely.nn.benchmark.semantic_segmentation.evaluator import (
    SemanticSegmentationEvalResult,
)
from supervisely.nn.task_type import TaskType
from supervisely.sly_logger import logger
from supervisely.task.progress import tqdm_sly

ComparisonVisualizer = Union[
    DetectionComparisonVisualizer, SemanticSegmentationComparisonVisualizer
]
ComparisonEvalResult = Union[ObjectDetectionEvalResult, SemanticSegmentationEvalResult]


class ModelComparison:
    """Orchestrates loading and visualizing comparison of multiple model evaluations."""

    def __init__(
        self,
        api: Api,
        remote_eval_dirs: List[str],
        progress: Optional[SlyTqdm] = None,
        workdir: Optional[str] = "./benchmark/comparison",
        cv_task: Optional[TaskType] = None,
        team_id: Optional[int] = None,
    ):
        """:param api: Supervisely API.
        :param remote_eval_dirs: Remote paths to evaluation result dirs.
        :param progress: Progress callback.
        :param workdir: Local work directory.
        :param cv_task: Optional task type override.
        :param team_id: Optional team ID.
        """
        self.api = api
        self.progress = progress or tqdm_sly
        self.workdir = workdir
        self.remote_eval_dirs = remote_eval_dirs
        self.eval_results: List[ComparisonEvalResult] = []
        self.task_type = cv_task
        self.team_id = team_id or env.team_id()

        eval_cls = SemanticSegmentationEvalResult
        eval_cls = ObjectDetectionEvalResult

        colors = get_predefined_colors(len(remote_eval_dirs) * 5)  # for better visualizations
        random.shuffle(colors)
        for i, eval_dir in enumerate(remote_eval_dirs):
            local_path = Path(self.workdir) / "eval_data" / Path(eval_dir).name
            self._load_eval_data(eval_dir, str(local_path))

            eval_cls = self._get_eval_cls(str(local_path))
            eval_result = eval_cls(local_path / "evaluation")
            eval_result.report_path = Path(eval_dir, "visualizations", "template.vue").as_posix()
            eval_result.color = rgb2hex(colors[i])

            self.eval_results.append(eval_result)

        self._validate_eval_data()

        self.visualizer: ComparisonVisualizer = None
        self.remote_dir = None

    def _validate_eval_data(self):
        """
        Validate the evaluation data before running the comparison.
        Make sure the benchmarks are done on the same project and datasets.
        """
        task_type = None
        img_names = None
        cat_names = None
        for eval_result in self.eval_results:
            next_task_type = eval_result.cv_task
            if not task_type is None:
                assert task_type == next_task_type, "Task types are different in the evaluations."
            task_type = next_task_type
            if task_type == TaskType.SEMANTIC_SEGMENTATION:
                next_img_names = set(eval_result.mp.per_image_metrics.index)
            else:
                next_img_names = set(
                    [img.get("file_name") for img in eval_result.coco_gt.imgs.values()]
                )
            if not img_names is None:
                assert img_names == next_img_names, "Images are different in the evaluations."
            img_names = next_img_names
            if task_type == TaskType.SEMANTIC_SEGMENTATION:
                next_cat_names = set(eval_result.mp.class_names)
            else:
                next_cat_names = set(eval_result.mp.cat_names)
            if not cat_names is None:
                assert cat_names == next_cat_names, "Categories are different in the evaluations."
            cat_names = next_cat_names

    def visualize(self):
        task_type = self.eval_results[0].cv_task
        if task_type in [
            TaskType.OBJECT_DETECTION,
            TaskType.INSTANCE_SEGMENTATION,
        ]:
            vis_cls = DetectionComparisonVisualizer
        elif task_type == TaskType.SEMANTIC_SEGMENTATION:
            vis_cls = SemanticSegmentationComparisonVisualizer
        else:
            raise ValueError(f"Unsupported task type: {task_type}")
        if self.visualizer is None:
            self.visualizer = vis_cls(self)
        self.visualizer.visualize()

    def upload_results(self, team_id: int, remote_dir: str, progress=None) -> str:
        self.remote_dir = self.visualizer.upload_results(team_id, remote_dir, progress)
        return self.remote_dir

    def get_report_link(self) -> str:
        if self.remote_dir is None:
            raise ValueError("Results are not uploaded yet.")
        return self.visualizer.renderer._get_report_link(self.api, self.team_id, self.remote_dir)

    @property
    def report(self):
        return self.visualizer.renderer.report

    @property
    def lnk(self):
        return self.visualizer.renderer.lnk

    def _load_eval_data(self, src_path: str, dst_path: str) -> None:
        dir_name = Path(src_path).name
        if not dir_empty(dst_path):
            logger.info(f"Directory {dst_path} is not empty. Skipping download.")
            return
        if not self.api.storage.dir_exists(self.team_id, src_path):
            raise ValueError(f"Directory {src_path} not found in storage.")
        mkdir(dst_path)
        with self.progress(
            message=f"Downloading evaluation data from {dir_name}",
            total=self.api.storage.get_directory_size(self.team_id, src_path),
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        ) as pbar:
            self.api.storage.download_directory(
                self.team_id, src_path, dst_path, progress_cb=pbar.update
            )

    def _get_cv_task(self, eval_dir: str) -> CVTask:
        try:
            eval_data = load_json_file(Path(eval_dir, "evaluation", "inference_info.json"))
            task_type = eval_data.get("task_type")
            return CVTask(task_type.replace(" ", "_").lower())
        except Exception as e:
            raise ValueError(
                f"Could not get CV task from `inference_info.json`, try to set it manually. {e}"
            )

    def _get_eval_cls(self, eval_dir: str) -> ComparisonEvalResult:
        if self.task_type is None:
            self.task_type = self._get_cv_task(eval_dir)
        if self.task_type in [
            CVTask.OBJECT_DETECTION,
            CVTask.INSTANCE_SEGMENTATION,
        ]:
            eval_cls = ObjectDetectionEvalResult
        elif self.task_type == CVTask.SEMANTIC_SEGMENTATION:
            eval_cls = SemanticSegmentationEvalResult
        else:
            raise ValueError(f"Unsupported task type: {self.task_type}")
        return eval_cls



# coding: utf-8
"""List and manage teams and monitor their activity."""

# docs
from __future__ import annotations

from typing import Callable, Dict, List, Literal, NamedTuple, Optional, Union

from tqdm import tqdm

from supervisely.api.module_api import ApiField, ModuleNoParent, UpdateableModule
from supervisely.sly_logger import logger


# @TODO - umar will add meta with review status and duration
class ActivityAction:
    """
    List of Team Actions to sort Team Activity.
    """

    LOGIN = "login"
    """"""
    LOGOUT = "logout"
    """"""
    CREATE_PROJECT = "create_project"
    """"""
    UPDATE_PROJECT = "update_project"
    """"""
    DISABLE_PROJECT = "disable_project"
    """"""
    RESTORE_PROJECT = "restore_project"
    """"""
    CREATE_DATASET = "create_dataset"
    """"""
    UPDATE_DATASET = "update_dataset"
    """"""
    DISABLE_DATASET = "disable_dataset"
    """"""
    RESTORE_DATASET = "restore_dataset"
    """"""
    CREATE_IMAGE = "create_image"
    """"""
    UPDATE_IMAGE = "update_image"
    """"""
    DISABLE_IMAGE = "disable_image"
    """"""
    RESTORE_IMAGE = "restore_image"
    """"""
    CREATE_FIGURE = "create_figure"
    """"""
    UPDATE_FIGURE = "update_figure"
    """"""
    DISABLE_FIGURE = "disable_figure"
    """"""
    RESTORE_FIGURE = "restore_figure"
    """"""
    CREATE_CLASS = "create_class"
    """"""
    UPDATE_CLASS = "update_class"
    """"""
    DISABLE_CLASS = "disable_class"
    """"""
    RESTORE_CLASS = "restore_class"
    """"""
    CREATE_BACKUP = "create_backup"
    """"""
    EXPORT_PROJECT = "export_project"
    """"""
    MODEL_TRAIN = "model_train"
    """"""
    MODEL_INFERENCE = "model_inference"
    """"""
    CREATE_PLUGIN = "create_plugin"
    """"""
    DISABLE_PLUGIN = "disable_plugin"
    """"""
    RESTORE_PLUGIN = "restore_plugin"
    """"""
    CREATE_NODE = "create_node"
    """"""
    DISABLE_NODE = "disable_node"
    """"""
    RESTORE_NODE = "restore_node"
    """"""
    CREATE_WORKSPACE = "create_workspace"
    """"""
    DISABLE_WORKSPACE = "disable_workspace"
    """"""
    RESTORE_WORKSPACE = "restore_workspace"
    """"""
    CREATE_MODEL = "create_model"
    """"""
    DISABLE_MODEL = "disable_model"
    """"""
    RESTORE_MODEL = "restore_model"
    """"""
    ADD_MEMBER = "add_member"
    """"""
    REMOVE_MEMBER = "remove_member"
    """"""
    LOGIN_TO_TEAM = "login_to_team"
    """"""
    ATTACH_TAG = "attach_tag"
    """"""
    UPDATE_TAG_VALUE = "update_tag_value"
    """"""
    DETACH_TAG = "detach_tag"
    """"""
    ANNOTATION_DURATION = "annotation_duration"
    """"""
    IMAGE_REVIEW_STATUS_UPDATED = "image_review_status_updated"
    """"""
    LABELER_ID = "labelerId"
    """"""
    REVIEW_STATUS = "reviewStatus"

    # case #1 - labeler pressed "finish image" button in labeling job
    # action: IMAGE_REVIEW_STATUS_UPDATED -> meta["reviewStatus"] == 'done'

    # case #2 - reviewer pressed "accept" or "reject" button
    # action: IMAGE_REVIEW_STATUS_UPDATED -> meta["reviewStatus"] == 'accepted' or 'rejected'

    # possible review statuses:
    # 'done' - i.e. labeler finished the image,
    # 'accepted' - reviewer
    # 'rejected' - reviewer

    # case #3 duration
    # action: ANNOTATION_DURATION -> meta["duration"] e.g. meta-> {"duration": 30} in seconds


class UsageInfo(NamedTuple):
    """Team usage/plan information returned by the API."""

    plan: Optional[str]


class TeamInfo(NamedTuple):
    """Team information returned by the API (basic fields + optional usage info)."""

    id: int
    name: str
    description: str
    role: str
    created_at: str
    updated_at: str
    usage: Optional[UsageInfo]


class TeamApi(ModuleNoParent, UpdateableModule):
    """API for working with teams."""

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                api = sly.Api.from_env()
                team_info = api.team.get_info_by_id(team_id)
        """
        ModuleNoParent.__init__(self, api)
        UpdateableModule.__init__(self, api)

    @staticmethod
    def info_sequence():
        """
        Sequence of fields that are returned by the API to represent TeamInfo.

        :Usage Example:

            .. code-block:: python

                TeamInfo(
                    id=1,
                    name='Vehicle',
                    description='',
                    role='admin',
                    created_at='2020-03-31T14:49:08.931Z',
                    updated_at='2020-03-31T14:49:08.931Z',
                    usage=UsageInfo(plan='free')
                )
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.DESCRIPTION,
            ApiField.ROLE,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.USAGE,
        ]

    @staticmethod
    def info_tuple_name():
        """
        Name of the tuple that represents TeamInfo.
        """
        return "TeamInfo"

    def get_list(self, filters: List[Dict[str, str]] = None) -> List[TeamInfo]:
        """
        List of all Teams on the Supervisely instance.

        :param filters: List of params to sort output Teams.
        :type filters: list, optional
        :returns: List of all Teams with information.
        :rtype: List[:class:`~supervisely.api.team_api.TeamInfo`]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                team_id = 8
                filters = [{"field": "name", "operator": "=", "value": "Team User123"}]
                team_list = api.team.get_list(filters=filters)
                print(team_list)
                # Output: [TeamInfo(id=1,
                #                   name='Vehicle',
                #                   description='',
                #                   role='admin',
                #                   created_at='2020-03-31T14:49:08.931Z',
                #                   updated_at='2020-03-31T14:49:08.931Z',
                #                   UsageInfo(plan='free')               ),
                # TeamInfo(id=2,
                #          name='Road',
                #          description='',
                #          role='admin',
                #          created_at='2020-03-31T08:52:11.000Z',
                #          updated_at='2020-03-31T08:52:11.000Z',
                #          UsageInfo(plan='free')               ),
                # TeamInfo(id=3,
                #          name='Animal',
                #          description='',
                #          role='admin',
                #          created_at='2020-04-02T08:59:03.717Z',
                #          updated_at='2020-04-02T08:59:03.717Z',
                #          UsageInfo(plan='free')               )
                # ]

                # Filtered Team list
                team_list = api.team.get_list(team_id, filters=[{ 'field': 'name', 'operator': '=', 'value': 'Animal' }])
                print(team_list)
                # Output: [TeamInfo(id=3,
                #                  name='Animal',
                #                  description='',
                #                  role='admin',
                #                  created_at='2020-04-02T08:59:03.717Z',
                #                  updated_at='2020-04-02T08:59:03.717Z',
                #                  UsageInfo(plan='free')               )
                # ]
        """
        return self.get_list_all_pages("teams.list", {ApiField.FILTER: filters or []})

    def get_info_by_id(self, id: int, raise_error: Optional[bool] = False) -> TeamInfo:
        """
        Get Team information by Team ID.

        :param id: Team ID in Supervisely.
        :type id: int
        :returns: Information about Team.
        :rtype: :class:`~supervisely.api.team_api.TeamInfo`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                team_info = api.team.get_info_by_id(8)
                print(team_info)
                # Output: TeamInfo(id=8,
                #          name='Fruits',
                #          description='',
                #          role='admin',
                #          created_at='2020-04-15T10:50:41.926Z',
                #          updated_at='2020-04-15T10:50:41.926Z',
                #          UsageInfo(plan='free')               )

                # You can also get Team info by name
                team_info = api.team.get_info_by_name("Fruits")
                print(team_info)
                # Output: TeamInfo(id=8,
                #          name='Fruits',
                #          description='',
                #          role='admin',
                #          created_at='2020-04-15T10:50:41.926Z',
                #          updated_at='2020-04-15T10:50:41.926Z',
                #          UsageInfo(plan='free')               )
        """

        info = self._get_info_by_id(id, "teams.info")
        if info is None and raise_error is True:
            raise KeyError(f"Team with id={id} not found in your account")
        return info

    def create(
        self,
        name: str,
        description: Optional[str] = "",
        change_name_if_conflict: Optional[bool] = False,
    ) -> TeamInfo:
        """
        Creates a new Team with the given name.

        :param name: Team name.
        :type name: str
        :param description: Team description.
        :type description: str
        :param change_name_if_conflict: Checks if given name already exists and adds suffix to the end of the name.
        :type change_name_if_conflict: bool, optional
        :returns: Information about Team.
        :rtype: :class:`~supervisely.api.team_api.TeamInfo`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                new_team = api.team.create("Flowers")
                print(new_team)
                # Output: TeamInfo(id=228,
                #                  name='Flowers',
                #                  description='',
                #                  role='admin',
                #                  created_at='2021-03-11T11:18:46.576Z',
                #                  updated_at='2021-03-11T11:18:46.576Z',
                #                  UsageInfo(plan='free')               )
        """
        effective_name = self._get_effective_new_name(
            name=name, change_name_if_conflict=change_name_if_conflict
        )
        response = self._api.post(
            "teams.add",
            {ApiField.NAME: effective_name, ApiField.DESCRIPTION: description},
        )
        return self._convert_json_info(response.json())

    def _get_update_method(self):
        """ """
        return "teams.editInfo"

    def get_activity(
        self,
        team_id: int,
        filter_user_id: Optional[int] = None,
        filter_project_id: Optional[int] = None,
        filter_job_id: Optional[int] = None,
        filter_actions: Optional[List] = None,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        labeler_id: Optional[int] = None,
        review_status: Optional[Literal["done", "accepted", "rejected"]] = None,
    ) -> List[Dict]:
        """
        Get Team activity by Team ID.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param filter_user_id: User ID by which the activity will be filtered.
        :type filter_user_id: int, optional
        :param filter_project_id: Project ID by which the activity will be filtered.
        :type filter_project_id: int, optional
        :param filter_job_id: Job ID by which the activity will be filtered.
        :type filter_job_id: int, optional
        :param filter_actions: List of ActivityAction by which the activity will be filtered.
        :type filter_actions: list, optional
        :param progress_cb: Function to check progress.
        :type progress_cb: tqdm or callable, optional
        :param start_date: Start date to get Team activity.
        :type start_date: str, optional
        :param end_date: End date to get Team activity.
        :type end_date: str, optional
        :param labeler_id: Labeler ID by which the activity will be filtered.
        :type labeler_id: int, optional
        :param review_status: Review status by which the activity will be filtered.
        :type review_status: str, optional, one of ['done', 'accepted', 'rejected']
        :returns: Team activity
        :rtype: :class:`List[dict]`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly
                from supervisely.api.team_api import ActivityAction as aa

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                labeling_actions = [
                    aa.ATTACH_TAG,
                    aa.UPDATE_TAG_VALUE,
                    aa.DETACH_TAG,
                ]

                team_activity = api.team.get_activity(8, filter_actions=labeling_actions)
                print(team_activity)
                # Output: [
                #     {
                #         "userId":7,
                #         "action":"detach_tag",
                #         "date":"2021-01-15T15:11:55.985Z",
                #         "user":"cxnt",
                #         "projectId":1817,
                #         "project":"App_Test_Poly",
                #         "datasetId":2370,
                #         "dataset":"train",
                #         "imageId":726985,
                #         "image":"IMG_8144.jpeg",
                #         "classId":"None",
                #         "class":"None",
                #         "figureId":"None",
                #         "job":"None",
                #         "jobId":"None",
                #         "tag":"hhhlk",
                #         "tagId":4720,
                #         "meta":{}
                #     },
                #     {
                #         "userId":7,
                #         "action":"attach_tag",
                #         "date":"2021-01-15T14:24:58.480Z",
                #         "user":"cxnt",
                #         "projectId":1817,
                #         "project":"App_Test_Poly",
                #         "datasetId":2370,
                #         "dataset":"train",
                #         "imageId":726985,
                #         "image":"IMG_8144.jpeg",
                #         "classId":"None",
                #         "class":"None",
                #         "figureId":"None",
                #         "job":"None",
                #         "jobId":"None",
                #         "tag":"hhhlk",
                #         "tagId":4720,
                #         "meta":{}
                #     }
                # ]
        """
        from datetime import datetime

        filters = []
        if labeler_id is not None:
            filters.append(
                {"field": ActivityAction.LABELER_ID, "operator": "=", "value": labeler_id}
            )
        if review_status is not None:
            filters.append(
                {
                    "field": ActivityAction.REVIEW_STATUS,
                    "operator": "=",
                    "value": review_status,
                }
            )

        if filter_user_id is not None:
            filters.append({"field": ApiField.USER_ID, "operator": "=", "value": filter_user_id})
        if filter_project_id is not None:
            filters.append(
                {
                    "field": ApiField.PROJECT_ID,
                    "operator": "=",
                    "value": filter_project_id,
                }
            )
        if filter_job_id is not None:
            filters.append({"field": ApiField.JOB_ID, "operator": "=", "value": filter_job_id})
        if filter_actions is not None:
            if type(filter_actions) is not list:
                raise TypeError(
                    "type(filter_actions) is {!r}. But has to be of type {!r}".format(
                        type(filter_actions), list
                    )
                )
            filters.append({"field": ApiField.TYPE, "operator": "in", "value": filter_actions})

        def _add_dt_filter(filters, dt, op):
            dt_iso = None
            if dt is None:
                return
            if type(dt) is str:
                dt_iso = dt
            elif type(dt) is datetime:
                dt_iso = dt.isoformat()
            else:
                raise TypeError(
                    "DT type must be string in ISO8601 format or datetime, not {}".format(type(dt))
                )
            filters.append({"field": ApiField.DATE, "operator": op, "value": dt_iso})

        _add_dt_filter(filters, start_date, ">=")
        _add_dt_filter(filters, end_date, "<=")

        method = "teams.activity"
        data = {ApiField.TEAM_ID: team_id, ApiField.FILTER: filters}
        first_response = self._api.post(method, data)
        first_response = first_response.json()

        total = first_response["total"]
        per_page = first_response["perPage"]
        pages_count = first_response["pagesCount"]
        results = first_response["entities"]

        def set_tqdm(progress_cb, results, total):
            progress_cb.total = total
            progress_cb.update(len(results) - progress_cb.n)
            progress_cb.refresh()

        if progress_cb is not None:
            if isinstance(progress_cb, tqdm):
                set_tqdm(progress_cb, results, total)
            else:
                progress_cb(len(results), total)
        if pages_count == 1 and len(first_response["entities"]) == total:
            pass
        else:
            for page_idx in range(2, pages_count + 1):
                temp_resp = self._api.post(method, {**data, "page": page_idx, "per_page": per_page})
                temp_items = temp_resp.json()["entities"]
                results.extend(temp_items)
                if progress_cb is not None:
                    if isinstance(progress_cb, tqdm):
                        set_tqdm(progress_cb, results, total)
                    else:
                        progress_cb(len(results), total)
            if len(results) != total:
                logger.warn(
                    f"Method '{method}': new events were created during pagination, "
                    f"downloaded={len(results)}, total={total}"
                )

        return results

    def _convert_json_info(self, info: dict, skip_missing=True) -> TeamInfo:
        """ """
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        res_dict = res._asdict()
        if isinstance(res_dict.get("usage"), dict):
            usage_dict = {f: res_dict["usage"].get(f) for f in UsageInfo._fields}
            res_dict["usage"] = UsageInfo(**usage_dict)
        return TeamInfo(**res_dict)



from supervisely.app.widgets import Widget, ConditionalWidget
from typing import Dict


class OneOf(Widget):
    """Renders exactly one of several child widgets based on conditional state (e.g. radio selection)."""

    def __init__(
        self,
        conditional_widget: ConditionalWidget,
        widget_id: str = None,
    ):
        """:param conditional_widget: Widget with conditional branches (e.g. radio + content map).
        :type conditional_widget: ConditionalWidget
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._conditional_widget = conditional_widget
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        return {}

    def get_json_state(self) -> Dict:
        return {}



from functools import wraps
from typing import Callable, Dict, List, Optional, Union

import yaml

import supervisely.app.widgets as Widgets
from supervisely.sly_logger import logger
from supervisely.task.progress import Progress


class ServingGUI:
    """GUI for model serving: device selection, serve/stop buttons, inference settings editor, download progress."""

    def __init__(self) -> None:
        self._device_select = Widgets.SelectCudaDevice(include_cpu_option=True)
        self._device_field = Widgets.Field(self._device_select, title="Device")
        self._serve_button = Widgets.Button("SERVE")
        self._success_label = Widgets.DoneLabel()
        self._success_label.hide()
        self._download_progress = Widgets.Progress("Downloading model...", hide_on_finish=True)
        self._download_progress.hide()
        self._change_model_button = Widgets.Button(
            "STOP AND CHOOSE ANOTHER MODEL", button_type="danger"
        )
        self._change_model_button.hide()

        self.serve_container = Widgets.Container(
            [
                self._device_field,
                self._download_progress,
                self._success_label,
                self._serve_button,
                self._change_model_button,
            ],
        )
        self.serve_model_card = Widgets.Card(
            title="Serve Model",
            description="Download and deploy the model on the selected device.",
            content=self.serve_container,
        )

        self._model_inference_settings_widget = Widgets.Editor(
            readonly=True, restore_default_button=False
        )
        self._model_inference_settings_container = Widgets.Field(
            self._model_inference_settings_widget,
            title="Inference settings",
            description="Model allows user to configure the following parameters on prediction phase",
        )

        self._model_info_widget = Widgets.ModelInfo()
        self._model_info_widget_container = Widgets.Field(
            self._model_info_widget,
            title="Session Info",
            description="Basic information about the deployed model",
        )

        self._model_classes_widget = Widgets.ClassesTable(selectable=False)
        self._model_classes_plug = Widgets.Text("No classes provided")
        self._model_classes_widget_container = Widgets.Field(
            content=Widgets.Container([self._model_classes_widget, self._model_classes_plug]),
            title="Model classes",
            description="List of classes model predicts",
        )

        self._model_full_info = Widgets.Container(
            [
                self._model_info_widget_container,
                self._model_inference_settings_container,
                self._model_classes_widget_container,
            ]
        )
        self._model_full_info.hide()
        self._before_deploy_msg = Widgets.Text("Deploy model to see the information.")

        self._model_full_info_card = Widgets.Card(
            title="Full model info",
            description="Inference settings, session parameters and model classes",
            collapsable=True,
            content=Widgets.Container(
                [
                    self._model_full_info,
                    self._before_deploy_msg,
                ]
            ),
        )

        self._model_full_info_card.collapse()
        self._additional_ui_content = []
        self.get_ui = self.__add_content_and_model_info_to_default_ui(
            self._model_full_info_card
        )  # pylint: disable=method-hidden

        self.on_change_model_callbacks: List[Callable] = [ServingGUI._hide_info_after_change]
        self.on_serve_callbacks: List[Callable] = []

        @self.serve_button.click
        def serve_model():
            self.deploy_with_current_params()

        @self._change_model_button.click
        def change_model():
            for cb in self.on_change_model_callbacks:
                cb(self)
            self.change_model()

    @property
    def serve_button(self) -> Widgets.Button:
        return self._serve_button

    @property
    def download_progress(self) -> Widgets.Progress:
        return self._download_progress

    @property
    def device(self) -> str:
        return self._device_select.get_device()

    def get_device(self) -> str:
        return self._device_select.get_device()

    def deploy_with_current_params(self):
        for cb in self.on_serve_callbacks:
            cb(self)
        self.set_deployed()

    def change_model(self):
        self._success_label.text = ""
        self._success_label.hide()
        self._serve_button.show()
        self._device_select._select.enable()
        self._device_select.enable()
        self._change_model_button.hide()
        # @TODO: Ask web team to add message to list of request ready messages
        # Progress("model deployment canceled", 1).iter_done_report()
        Progress("Application is started ...", 1).iter_done_report()

    def _hide_info_after_change(self):
        self._model_full_info_card.collapse()
        self._model_full_info.hide()
        self._before_deploy_msg.show()

    def set_deployed(self, device: str = None):
        if device is not None:
            self._device_select.set_device(device)
        self._success_label.text = f"Model has been successfully loaded on {self._device_select.get_device().upper()} device"
        self._success_label.show()
        self._serve_button.hide()
        self._device_select._select.disable()
        self._device_select.disable()
        self._change_model_button.show()
        Progress("Model deployed", 1).iter_done_report()

    def show_deployed_model_info(self, inference):
        self.set_inference_settings(inference)
        self.set_project_meta(inference)
        self.set_model_info(inference)
        self._before_deploy_msg.hide()
        self._model_full_info.show()
        self._model_full_info_card.uncollapse()

    def set_inference_settings(self, inference):
        if len(inference.custom_inference_settings_dict.keys()) == 0:
            inference_settings_str = "# inference settings dict is empty"
        else:
            inference_settings_str = yaml.dump(inference.custom_inference_settings_dict)
        self._model_inference_settings_widget.set_text(inference_settings_str, "yaml")
        self._model_inference_settings_widget.show()

    def set_project_meta(self, inference):
        if self._get_classes_from_inference(inference) is None:
            logger.warn("Skip loading project meta.")
            self._model_classes_widget.hide()
            self._model_classes_plug.show()
            return

        self._model_classes_widget.set_project_meta(inference.model_meta)
        self._model_classes_plug.hide()
        self._model_classes_widget.show()

    def set_model_info(self, inference):
        info = inference.get_human_readable_info(replace_none_with="Not provided")
        self._model_info_widget.set_model_info(inference.task_id, info)

    def _get_classes_from_inference(self, inference) -> Optional[List[str]]:
        classes = None
        try:
            classes = inference.get_classes()
        except NotImplementedError:
            logger.warn(f"get_classes() function not implemented for {type(inference)} object.")
        except AttributeError:
            logger.warn("Probably, get_classes() function not working without model deploy.")
        except Exception as exc:
            logger.warn("Skip getting classes info due to exception")
            logger.exception(exc)

        if classes is None or len(classes) == 0:
            logger.warn(f"get_classes() function return {classes}; skip classes processing.")
            return None
        return classes

    def get_ui(self) -> Widgets.Widget:  # pylint: disable=method-hidden
        return Widgets.Container([self.serve_model_card])

    def add_content_to_default_ui(
        self, widgets: Union[Widgets.Widget, List[Widgets.Widget]]
    ) -> None:
        if isinstance(widgets, List):
            self._additional_ui_content.extend(widgets)
        else:
            self._additional_ui_content.append(widgets)

    def __add_content_and_model_info_to_default_ui(
        self,
        model_info_widget: Widgets.Widget,
    ) -> Callable:
        def decorator(get_ui):
            @wraps(get_ui)
            def wrapper(*args, **kwargs):
                ui = get_ui(*args, **kwargs)
                content = [ui, *self._additional_ui_content, model_info_widget]
                ui_with_info = Widgets.Container(content)
                return ui_with_info

            return wrapper

        return decorator(self.get_ui)



# coding: utf-8
"""Create and manage Agents in Supervisely."""

from __future__ import annotations

from enum import Enum
from typing import Dict, List, Literal, NamedTuple, Optional

from supervisely.api.module_api import ApiField, ModuleApi, ModuleWithStatus


class AgentInfo(NamedTuple):
    """
    NamedTuple with agent information.
    """

    id: int
    name: str
    token: str
    status: str
    user_id: int
    team_id: int
    capabilities: dict
    created_at: str
    updated_at: str
    is_public: Optional[bool]
    type: Optional[str]
    disabled: Optional[bool]
    version: Optional[str]
    gpu_info: Optional[dict]


class AgentNotFound(Exception):
    """Agent was not found."""

    pass


class AgentNotRunning(Exception):
    """Agent is not running."""

    pass


class AgentApi(ModuleApi, ModuleWithStatus):
    """API for working with agents."""

    class Status(Enum):
        """Agent runtime status values returned by the platform."""

        WAITING = "waiting"
        """"""
        RUNNING = "running"
        """"""

    @staticmethod
    def info_sequence():
        """
        NamedTuple AgentInfo information about Agent.

        :Usage Example:

            .. code-block:: python

                AgentInfo(
                    id=1,
                    name="Agent",
                    token="***",
                    status="running",
                    user_id=1,
                    team_id=1,
                    capabilities={},
                )
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.TOKEN,
            ApiField.STATUS,
            ApiField.USER_ID,
            ApiField.TEAM_ID,
            ApiField.CAPABILITIES,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.IS_PUBLIC,
            ApiField.TYPE,
            ApiField.DISABLED,
            ApiField.VERSION,
            ApiField.GPU_INFO,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **AgentInfo**.
        """
        return "AgentInfo"

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                api = sly.Api.from_env()
                agents = api.agent.get_list(team_id=8)
        """
        ModuleApi.__init__(self, api)
        ModuleWithStatus.__init__(self)

    def get_list(
        self, team_id: int, filters: Optional[List[Dict[str, str]]] = None
    ) -> List[NamedTuple]:
        """
        List of all agents in the given Team.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param filters: List of params to sort output Agents.
        :type filters: List[dict], optional
        :returns: List of agents with information. See :meth:`~supervisely.api.agent_api.AgentApi.info_sequence`.
        :rtype: List[NamedTuple]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                team_id = 16087
                agents = api.agent.get_list(team_id)

                filter_agents = api.agent.get_list(team_id, filters=[{ 'field': 'name', 'operator': '=', 'value': 'Gorgeous Chicken' }])
        """
        return self.get_list_all_pages("agents.list", {"teamId": team_id, "filter": filters or []})

    def get_list_available(
        self,
        team_id: int,
        show_public: bool = False,
        show_only_running: bool = True,
        has_gpu: bool = False,
        type: Literal[
            "import",
            "import_agent",
            "export",
            "train",
            "inference",
            "infer_rpc",
            "custom",
            "pipeline",
            "python",
            "app",
            "app_gpu",
            "smarttool",
        ] = None,
        plugin_version_id: int = None,
        version: str = None,
        envs: List[dict] = None,
        min_nvidia_driver_version: str = None,
    ) -> List[AgentInfo]:
        """
        Return list of available agents. Available agents are agents that are not disabled and can be used to run the app.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param show_public: Show public agents.
        :type show_public: bool, optional
        :param show_only_running: Show only running agents.
        :type show_only_running: bool, optional
        :param has_gpu: Show only agents with GPU.
        :type has_gpu: bool, optional
        :param type: Filter by agent task type.
        :type type: Literal["import", "import_agent", "export", "train", "inference", "infer_rpc", "custom", "pipeline", "python", "app", "app_gpu", "smarttool"], optional
        :param plugin_version_id: Filter by agent plugin version ID.
        :type plugin_version_id: int, optional
        :param version: Filter by agent version.
        :type version: str, optional
        :param envs: Filter by agent env variables.
        :type envs: List[dict], optional
        :param min_nvidia_driver_version: Filter by minimum nvidia driver version.
        :type min_nvidia_driver_version: str, optional
        :returns: List of agents with information. See :class:`~supervisely.api.agent_api.AgentInfo`.
        :rtype: List[:class:`~supervisely.api.agent_api.AgentInfo`]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                team_id = 350
                available_agents = api.agent.get_list_available(
                    team_id=team_id,
                    show_public=True,
                    show_only_running=False,
                    has_gpu=False,
                    type="app",
                    plugin_version_id=1,
                    version="6.7.39",
                    envs=[{"field": "DOCKER_NET"}],
                    min_nvidia_driver_version="546.33"
                )
        """
        data = {
            "teamId": team_id,
            "withoutPublic": not show_public,
            "anyStatus": not show_only_running,
            "needGPU": has_gpu,
        }

        if type is not None:
            data["type"] = type
        if plugin_version_id is not None:
            data["pluginVersionId"] = plugin_version_id
        if version is not None:
            data["version"] = version
        if envs is not None:
            data["envs"] = envs
        if min_nvidia_driver_version is not None:
            data["minNvidiaDriverVersion"] = min_nvidia_driver_version
        response = self._api.post("agents.available", data)
        infos = response.json()
        agent_infos = []
        for info in infos:
            agent_infos.append(self._convert_json_info(info, True))
        return agent_infos

    def get_info_by_id(self, id: int) -> NamedTuple:
        """
        Get Agent information by ID.

        :param id: Agent ID in Supervisely.
        :type id: int
        :returns: Information about an agent. See :meth:`~supervisely.api.agent_api.AgentApi.info_sequence`.
        :rtype: NamedTuple

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                agent = api.agent.get_info_by_id(7)
        """
        return self._get_info_by_id(id, "agents.info")

    def get_status(self, id: int) -> AgentApi.Status:
        """
        Status object containing status of Agent: waiting or running.

        :param id: Agent ID in Supervisely.
        :type id: int
        :returns: Agent Status
        :rtype: :class:`~supervisely.api.agent_api.AgentApi.Status`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                agent = api.agent.get_status(7)
        """
        status_str = self.get_info_by_id(id).status
        return self.Status(status_str)

    def raise_for_status(self, status):
        """raise_for_status"""
        pass

    def _convert_json_info(self, info: dict, skip_missing=False):
        return super()._convert_json_info(info, True)



# coding: utf-8

"""Work with point cloud objects via the Supervisely API."""

from typing import List

from supervisely.api.entity_annotation.object_api import ObjectApi
from supervisely.api.pointcloud.pointcloud_tag_api import PointcloudObjectTagApi
from supervisely.pointcloud_annotation.pointcloud_object_collection import (
    PointcloudObjectCollection,
)
from supervisely.video_annotation.key_id_map import KeyIdMap


class PointcloudObjectApi(ObjectApi):
    """
    API for working with point cloud objects.
    """

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`
        """
        super().__init__(api)
        self.tag = PointcloudObjectTagApi(api)

    def append_bulk(
        self,
        pointcloud_id: int,
        objects: PointcloudObjectCollection,
        key_id_map: KeyIdMap = None,
    ) -> List[int]:
        """
        Add point cloud objects to annotation objects.

        :param pointcloud_id: Point cloud ID in Supervisely.
        :type pointcloud_id: int
        :param objects: PointcloudObjectCollection objects collection.
        :type objects: :class:`~supervisely.pointcloud_annotation.pointcloud_object_collection.PointcloudObjectCollection`
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: List of object IDs
        :rtype: :class:`List[int]`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly
                from supervisely.video_annotation.key_id_map import KeyIdMap

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                project_id = 19442
                pointcloud_id = 19618685

                meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(meta_json)

                key_id_map = KeyIdMap()
                ann_info = api.pointcloud.annotation.download(pointcloud_id)
                ann = sly.PointcloudAnnotation.from_json(ann_info, project_meta, key_id_map)

                res = api.pointcloud.object.append_bulk(pointcloud_id, ann.objects, key_id_map)
                print(res)

                # Output: [5565915, 5565916, 5565917, 5565918, 5565919]
        """

        info = self._api.pointcloud.get_info_by_id(pointcloud_id)
        return self._append_bulk(
            self._api.pointcloud.tag,
            pointcloud_id,
            info.project_id,
            info.dataset_id,
            objects,
            key_id_map,
            is_pointcloud=True,
        )

    def append_to_dataset(
        self,
        dataset_id: int,
        objects: PointcloudObjectCollection,
        key_id_map: KeyIdMap = None,
    ) -> List[int]:
        """
        Add pointcloud objects to Dataset annotation objects.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param objects: PointcloudObjectCollection objects collection.
        :type objects: :class:`~supervisely.pointcloud_annotation.pointcloud_object_collection.PointcloudObjectCollection`
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: List of objects IDs
        :rtype: :class:`List[int]`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly
                from supervisely.pointcloud_annotation.pointcloud_annotation import PointcloudObjectCollection
                from supervisely.video_annotation.key_id_map import KeyIdMap

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                project_id = 19442
                project = api.project.get_info_by_id(project_id)
                dataset = api.dataset.create(project.id, "demo_dataset", change_name_if_conflict=True)

                class_car = sly.ObjClass('car', sly.Cuboid)
                class_pedestrian = sly.ObjClass('pedestrian', sly.Cuboid)
                classes = sly.ObjClassCollection([class_car, class_pedestrian])
                project_meta = sly.ProjectMeta(classes)
                updated_meta = api.project.update_meta(project.id, project_meta.to_json())

                key_id_map = KeyIdMap()

                pedestrian_object = sly.PointcloudObject(class_pedestrian)
                car_object = sly.PointcloudObject(class_car)
                objects_collection = PointcloudObjectCollection([pedestrian_object, car_object])

                uploaded_objects_ids = api.pointcloud_episode.object.append_to_dataset(
                    dataset.id,
                    objects_collection,
                    key_id_map,
                )
                print(uploaded_objects_ids)

                # Output: [5565920, 5565921, 5565922]
        """

        project_id = self._api.dataset.get_info_by_id(dataset_id).project_id
        return self._append_bulk(
            self._api.pointcloud.tag,
            dataset_id,
            project_id,
            dataset_id,
            objects,
            key_id_map,
            is_pointcloud=True,
        )



import numpy as np
import cv2

class SegmentationMetrics:
    """Computes segmentation metrics: mIoU, boundary IoU, pixel accuracy with optional ignore index."""

    def __init__(
        self,
        num_classes: int,
        ignore_index: int = 255,
        boundary_dilation_ratio: float = 0.02,
    ):
        """
        :param num_classes: Number of classes.
        :type num_classes: int
        :param ignore_index: Ignored label.
        :type ignore_index: int
        :param boundary_dilation_ratio: Boundary dilation for boundary IoU.
        :type boundary_dilation_ratio: float
        """
        self.num_classes = num_classes
        self.ignore_index = ignore_index
        self.boundary_dilation_ratio = boundary_dilation_ratio

    def _boundary_iou(self, gt: np.ndarray, pred: np.ndarray) -> float:
        gt_b = self._mask_to_boundary(gt)
        pred_b = self._mask_to_boundary(pred)

        inter = ((gt_b & pred_b) > 0).sum()
        union = ((gt_b | pred_b) > 0).sum()

        return inter / union if union > 0 else 0.0

    def _mask_to_boundary(self, mask: np.ndarray) -> np.ndarray:
        h, w = mask.shape
        diag = np.sqrt(h * h + w * w)
        dilation = max(1, int(round(self.boundary_dilation_ratio * diag)))

        padded = cv2.copyMakeBorder(mask, 1, 1, 1, 1, cv2.BORDER_CONSTANT, value=0)

        kernel = np.ones((3, 3), np.uint8)
        eroded = cv2.erode(padded, kernel, iterations=dilation)
        eroded = eroded[1 : h + 1, 1 : w + 1]

        return mask - eroded

    def all_metrics(
        self,
        pred: np.ndarray,
        gt: np.ndarray,
        exclude_background: bool = True,
    ) -> dict:
        if pred.shape != gt.shape:
            raise ValueError("Shape mismatch")

        valid = gt != self.ignore_index
        start = 1 if exclude_background else 0

        per_class_iou = {}
        per_class_boundary_iou = {}

        for cls in range(start, self.num_classes):
            gt_cls = (gt == cls) & valid
            if not gt_cls.any():
                continue

            pred_cls = (pred == cls) & valid

            gt_bin = gt_cls.astype(np.uint8)
            pred_bin = pred_cls.astype(np.uint8)

            inter = np.logical_and(gt_cls, pred_cls).sum()
            union = np.logical_or(gt_cls, pred_cls).sum()
            per_class_iou[cls] = inter / union if union > 0 else 0.0

            per_class_boundary_iou[cls] = self._boundary_iou(gt_bin, pred_bin)

        mean_iou = float(np.mean(list(per_class_iou.values()))) if per_class_iou else 0.0
        mean_boundary_iou = (
            float(np.mean(list(per_class_boundary_iou.values())))
            if per_class_boundary_iou else 0.0
        )

        return {
            "mean_iou": mean_iou,
            "mean_boundary_iou": mean_boundary_iou,
            "per_class_iou": per_class_iou,
            "per_class_boundary_iou": per_class_boundary_iou,
        }



# isort: skip_file

import copy
import io

import numpy as np
import pandas as pd
import re
from typing import NamedTuple, Any, List, Optional, Dict

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from supervisely.app import DataJson
from supervisely.app.content import StateJson
from supervisely.app.widgets import Widget
from supervisely.sly_logger import logger
import traceback
from fastapi.responses import StreamingResponse, RedirectResponse


class PackerUnpacker:
    """Helpers to (un)pack supported table input types (dict, pandas DataFrame) into a common JSON shape."""

    SUPPORTED_TYPES = tuple([dict, pd.DataFrame])

    @staticmethod
    def validate_sizes(unpacked_data):
        for row in unpacked_data["data"]:
            if len(row) != len(unpacked_data["columns"]):
                raise ValueError(
                    "Sizes mismatch:\n"
                    f'{len(row)} != {len(unpacked_data["columns"])}\n'
                    f"{row}\n"
                    f'{unpacked_data["columns"]}'
                )

    @staticmethod
    def unpack_data(data, unpacker_cb):
        unpacked_data = unpacker_cb(data)
        PackerUnpacker.validate_sizes(unpacked_data)
        return unpacked_data

    @staticmethod
    def pack_data(data, packer_cb):
        packed_data = packer_cb(data)
        return packed_data

    @staticmethod
    def dict_unpacker(data: dict):
        unpacked_data = copy.deepcopy(data)
        if "summaryRow" not in unpacked_data.keys():
            unpacked_data["summaryRow"] = None

        return unpacked_data

    @staticmethod
    def pandas_unpacker(data: pd.DataFrame):
        # Keep None/NaN values in source data, don't replace them
        # They will be converted to "" only when sending to frontend
        unpacked_data = {
            "columns": data.columns.to_list(),
            "data": data.values.tolist(),
            "summaryRow": None,
        }
        return unpacked_data

    @staticmethod
    def dict_packer(data):
        packed_data = {"columns": data["columns"], "data": data["data"]}
        if "summaryRow" in data.keys() and data["summaryRow"] is not None:
            packed_data["summaryRow"] = data["summaryRow"]
        return packed_data

    @staticmethod
    def pandas_packer(data):
        packed_data = pd.DataFrame(data=data["data"], columns=data["columns"])
        return packed_data


DATATYPE_TO_PACKER = {
    pd.DataFrame: PackerUnpacker.pandas_packer,
    dict: PackerUnpacker.dict_packer,
}

DATATYPE_TO_UNPACKER = {
    pd.DataFrame: PackerUnpacker.pandas_unpacker,
    dict: PackerUnpacker.dict_unpacker,
}

"""
iris = pd.read_csv(
    "https://raw.githubusercontent.com/mwaskom/seaborn-data/master/iris.csv"
)
iris.insert(loc=0, column="index", value=np.arange(len(iris)))
table = sly.app.widgets.Table(data=iris, fixed_cols=1)

@table.click
def show_image(datapoint: sly.app.widgets.Table.ClickedDataPoint):
    print("Column name = ", datapoint.column_name)
    print("Cell value = ", datapoint.cell_value)
    print("Row = ", datapoint.row)

"""


class Table(Widget):
    """Table widget for displaying tabular data with sorting, pagination and click callbacks."""

    class Routes:
        """HTTP routes used by the widget frontend for callbacks/download."""

        CELL_CLICKED = "cell_clicked_cb"
        DOWNLOAD_AS_CSV = "download_as_csv"
        UPDATE_SORT = "update_sort_cb"

    class ClickedDataPoint:
        """Payload describing a clicked table cell and its row context."""

        def __init__(self, column_index: int, column_name: str, cell_value: Any, row: dict, row_index: int = None):
            """:param column_index: 0-based column index.
            :type column_index: int
            :param column_name: Column name.
            :type column_name: str
            :param cell_value: Cell value.
            :type cell_value: Any
            :param row: Full row as dict.
            :type row: dict
            :param row_index: 0-based row index.
            :type row_index: int, optional
            """
            self.column_index = column_index
            self.column_name = column_name
            self.cell_value = cell_value
            self.row = row
            self.row_index = row_index
            self.button_name = None
            search = re.search(r"<button>(.*?)</button>", str(self.cell_value))
            if search is not None:
                self.button_name = search.group(1)

    def __init__(
        self,
        data=None,
        columns: Optional[list] = None,
        fixed_cols: Optional[int] = None,
        per_page: Optional[int] = 10,
        page_sizes: Optional[List[int]] = [10, 15, 30, 50, 100],
        width: Optional[str] = "auto",  # "200px", or "100%"
        sort_column_id: int = 0,
        sort_direction: Optional[Literal["asc", "desc"]] = "asc",
        widget_id: Optional[str] = None,
    ):
        """
        :param data: Data of table in different formats:
        1. Pandas Dataframe or pd.DataFrame(data=data, columns=columns)
        2. Python dict with structure {
                                        'columns': ['col_name_1', 'col_name_2', ...],
                                        'data': [
                                                    ['row_1_column_1', 'row_1_column_2', ...],
                                                    ['row_2_column_1', 'row_2_column_2', ...],
                                                    ...
                                                ]
                                      }
        """

        self._supported_types = PackerUnpacker.SUPPORTED_TYPES

        self._parsed_data = None
        self._data_type = None
        self._click_handled = False

        self._update_table_data(input_data=pd.DataFrame(data=data, columns=columns))

        self._per_page = per_page
        self._page_sizes = page_sizes
        self._fix_columns = fixed_cols
        self._width = width
        self._loading = False

        self._sort_column_id = (
            sort_column_id if self._validate_sort(column_id=sort_column_id) else 0
        )
        self._sort_direction = (
            sort_direction if self._validate_sort(direction=sort_direction) else "asc"
        )

        super().__init__(widget_id=widget_id, file_path=__file__)

    def _prepare_data_for_frontend(self, data_dict):
        """Convert None and NaN values to empty strings for frontend display.
        This preserves the original None/NaN values in _parsed_data.
        """
        import math

        display_data = copy.deepcopy(data_dict)

        # Convert None/NaN in data rows
        for row in display_data.get("data", []):
            for i in range(len(row)):
                value = row[i]
                # Check for None or NaN (NaN is a float that doesn't equal itself)
                if value is None or (isinstance(value, float) and math.isnan(value)):
                    row[i] = ""

        # Convert None/NaN in summary row if present
        if "summaryRow" in display_data and display_data["summaryRow"] is not None:
            summary_row = display_data["summaryRow"]
            for i in range(len(summary_row)):
                value = summary_row[i]
                if value is None or (isinstance(value, float) and math.isnan(value)):
                    summary_row[i] = ""

        return display_data

    def get_json_data(self):
        return {
            "table_data": self._prepare_data_for_frontend(self._parsed_data),
            "table_options": {
                "perPage": self._per_page,
                "pageSizes": self._page_sizes,
                "fixColumns": self._fix_columns,
                "customDownloadUrl": None,
            },
            "loading": self._loading,
        }

    def get_json_state(self):
        return {
            "selected_row": {},
            "sort": {
                "columnIdx": self._sort_column_id,
                "direction": self._sort_direction,
            },
        }

    def _update_table_data(self, input_data):
        if input_data is not None:
            self._parsed_data = copy.deepcopy(self._get_unpacked_data(input_data=input_data))
        else:
            self._parsed_data = {"columns": [], "data": [], "summaryRow": None}
            self._data_type = dict

    def _get_packed_data(self, input_data, data_type):
        return PackerUnpacker.pack_data(data=input_data, packer_cb=DATATYPE_TO_PACKER[data_type])

    def _get_unpacked_data(self, input_data):
        input_data_type = type(input_data)

        if input_data_type not in self._supported_types:
            raise TypeError(
                f"Cannot parse input data, please use one of supported datatypes: {self._supported_types}\n"
                """
                            1. Pandas Dataframe \n
                            2. Python dict with structure {
                                        'columns': ['col_name_1', 'col_name_2', ...],
                                        'data': [
                                                            ['row_1_column_1', 'row_1_column_2', ...],
                                                            ['row_2_column_1', 'row_2_column_2', ...],
                                                            ...
                                                          ]
                                      }
                            """
            )

        return PackerUnpacker.unpack_data(
            data=input_data, unpacker_cb=DATATYPE_TO_UNPACKER[input_data_type]
        )

    @property
    def fixed_columns_num(self) -> int:
        return self._fix_columns

    @fixed_columns_num.setter
    def fixed_columns_num(self, value: int):
        self._fix_columns = value
        DataJson()[self.widget_id]["table_options"]["fixColumns"] = self._fix_columns

    @property
    def summary_row(self) -> List[Any]:
        if "summaryRow" not in self._parsed_data.keys():
            return None
        return self._parsed_data["summaryRow"]

    @summary_row.setter
    def summary_row(self, value: List[Any]):
        cols_num = len(self._parsed_data["columns"])
        if len(value) < cols_num:
            value.extend([""] * (cols_num - len(value)))
        elif len(value) > cols_num:
            value = value[:cols_num]
        DataJson()[self.widget_id]["table_data"]["summaryRow"] = value

    def to_json(self) -> Dict:
        return self._get_packed_data(self._parsed_data, dict)

    def to_pandas(self) -> pd.DataFrame:
        return self._get_packed_data(self._parsed_data, pd.DataFrame)

    def read_json(self, value: dict) -> None:
        self._update_table_data(input_data=value)
        DataJson()[self.widget_id]["table_data"] = self._prepare_data_for_frontend(
            self._parsed_data
        )
        DataJson().send_changes()
        self.clear_selection()

    def read_pandas(self, value: pd.DataFrame) -> None:
        self._update_table_data(input_data=value)
        DataJson()[self.widget_id]["table_data"] = self._prepare_data_for_frontend(
            self._parsed_data
        )
        DataJson().send_changes()
        self.clear_selection()

    def insert_row(self, data, index=-1):
        PackerUnpacker.validate_sizes({"columns": self._parsed_data["columns"], "data": [data]})

        table_data = self._parsed_data["data"]
        index = len(table_data) if index > len(table_data) or index < 0 else index

        self._parsed_data["data"].insert(index, data)
        DataJson()[self.widget_id]["table_data"] = self._prepare_data_for_frontend(
            self._parsed_data
        )
        DataJson().send_changes()

    def pop_row(self, index=-1):
        index = (
            len(self._parsed_data["data"]) - 1
            if index > len(self._parsed_data["data"]) or index < 0
            else index
        )

        if len(self._parsed_data["data"]) != 0:
            popped_row = self._parsed_data["data"].pop(index)
            DataJson()[self.widget_id]["table_data"] = self._prepare_data_for_frontend(
                self._parsed_data
            )
            DataJson().send_changes()
            return popped_row

    def get_selected_cell(self, state):
        # logger.debug(
        #     "Selected row",
        #     extra={"selected_row": state[self.widget_id]["selected_row"]},
        # )
        row_index = state[self.widget_id]["selected_row"].get("selectedRow")
        column_name = state[self.widget_id]["selected_row"].get("selectedColumnName")
        column_index = state[self.widget_id]["selected_row"].get("selectedColumn")
        row = state[self.widget_id]["selected_row"].get("selectedRowData")

        if row_index is None or column_name is None or column_index is None or row is None:
            # click table header or clear selection
            return None

        return {
            "column_index": column_index,
            "column_name": column_name,
            "row": row,
            "row_index": row_index,
            "cell_value": row[column_name],
        }

    def download_as_csv(self, func):
        route_path = self.get_route_path(Table.Routes.DOWNLOAD_AS_CSV)
        server = self._sly_app.get_server()
        DataJson()[self.widget_id]["table_options"]["customDownloadUrl"] = route_path

        @server.get(route_path)
        def _click():
            try:
                df = func()
                df_name = df.index.name
                if df_name is None:
                    df_name = "table"
                stream = io.StringIO()
                df.to_csv(stream, index=False)
                response = StreamingResponse(iter([stream.getvalue()]), media_type="text/csv")
                response.headers["Content-Disposition"] = f"attachment; filename={df_name}.csv"
                return response

            except Exception as e:
                logger.error(traceback.format_exc(), exc_info=True, extra={"exc_str": str(e)})
                raise e

        return _click

    def click(self, func):
        route_path = self.get_route_path(Table.Routes.CELL_CLICKED)
        update_sorting_column_route_path = self.get_route_path(Table.Routes.UPDATE_SORT)
        server = self._sly_app.get_server()

        @server.post(update_sorting_column_route_path)
        def _update_sorting_column():
            StateJson().send_changes()
            return

        self._click_handled = True

        @server.post(route_path)
        def _click():
            try:
                value_dict = self.get_selected_cell(StateJson())
                if value_dict is None:
                    return
                datapoint = Table.ClickedDataPoint(**value_dict)
                func(datapoint)
            except Exception as e:
                logger.error(traceback.format_exc(), exc_info=True, extra={"exc_str": str(e)})
                raise e

        return _click

    @staticmethod
    def get_html_text_as_button(title="preview"):
        return f"<button>{title}</button>"

    @staticmethod
    def create_button(title) -> str:
        return f"<button>{title}</button>"

    @property
    def loading(self):
        return self._loading

    @loading.setter
    def loading(self, value: bool):
        self._loading = value
        DataJson()[self.widget_id]["loading"] = self._loading
        DataJson().send_changes()

    def clear_selection(self):
        StateJson()[self.widget_id]["selected_row"] = {}
        StateJson().send_changes()

    @staticmethod
    def _values_equal(val1, val2):
        """Compare two values, handling NaN specially."""
        import math

        # Check if both are NaN
        is_nan1 = isinstance(val1, float) and math.isnan(val1)
        is_nan2 = isinstance(val2, float) and math.isnan(val2)
        if is_nan1 and is_nan2:
            return True
        # Check if both are None
        if val1 is None and val2 is None:
            return True
        # Regular comparison
        return val1 == val2

    def delete_row(self, key_column_name, key_cell_value):
        col_index = self._parsed_data["columns"].index(key_column_name)
        row_indices = []
        for idx, row in enumerate(self._parsed_data["data"]):
            if self._values_equal(row[col_index], key_cell_value):
                row_indices.append(idx)
        if len(row_indices) == 0:
            raise ValueError('Column "{key_column_name}" does not have value "{key_cell_value}"')
        if len(row_indices) > 1:
            raise ValueError(
                'Column "{key_column_name}" has multiple cells with the value "{key_cell_value}". Value has to be unique'
            )
        self.pop_row(row_indices[0])

    def update_cell_value(self, key_column_name, key_cell_value, column_name, new_value):
        key_col_index = self._parsed_data["columns"].index(key_column_name)
        row_indices = []
        for idx, row in enumerate(self._parsed_data["data"]):
            if self._values_equal(row[key_col_index], key_cell_value):
                row_indices.append(idx)
        if len(row_indices) == 0:
            raise ValueError('Column "{key_column_name}" does not have value "{key_cell_value}"')
        if len(row_indices) > 1:
            raise ValueError(
                'Column "{key_column_name}" has multiple cells with the value "{key_cell_value}". Value has to be unique'
            )

        col_index = self._parsed_data["columns"].index(column_name)
        self._parsed_data["data"][row_indices[0]][col_index] = new_value
        DataJson()[self.widget_id]["table_data"] = self._prepare_data_for_frontend(
            self._parsed_data
        )
        DataJson().send_changes()

    def update_matching_cells(self, key_column_name, key_cell_value, column_name, new_value):
        key_col_index = self._parsed_data["columns"].index(key_column_name)
        row_indices = []
        for idx, row in enumerate(self._parsed_data["data"]):
            if self._values_equal(row[key_col_index], key_cell_value):
                row_indices.append(idx)

        col_index = self._parsed_data["columns"].index(column_name)
        for row_idx in row_indices:
            self._parsed_data["data"][row_idx][col_index] = new_value
        DataJson()[self.widget_id]["table_data"] = self._prepare_data_for_frontend(
            self._parsed_data
        )
        DataJson().send_changes()

    def sort(self, column_id: int = None, direction: Optional[Literal["asc", "desc"]] = None):
        self._validate_sort(column_id, direction)
        if column_id is not None:
            self._sort_column_id = column_id
            StateJson()[self.widget_id]["sort"]["columnIdx"] = column_id
        if direction is not None:
            self._sort_direction = direction
            StateJson()[self.widget_id]["sort"]["direction"] = direction
        StateJson().send_changes()

    def _validate_sort(
        self, column_id: int = None, direction: Optional[Literal["asc", "desc"]] = None
    ):
        if column_id is not None and type(column_id) is not int:
            raise ValueError(f'Incorrect value of "column_id": {type(column_id)} is not "int".')
        if column_id is not None and column_id < 0:
            raise ValueError(f'Incorrect value of "column_id": {column_id} < 0')
        if direction is not None and direction not in ["asc", "desc"]:
            raise ValueError(
                f'Incorrect value of "direction": {direction}. Value can be one of "asc" or "desc".'
            )
        return True



import mimetypes
import re
from pathlib import Path
from typing import List, Union

import magic
import numpy as np
from PIL import Image

from supervisely import Label, Rectangle, logger
from supervisely.geometry.oriented_bbox import OrientedBBox
from supervisely.imaging.image import read, write
from supervisely.io.fs import (
    get_file_ext,
    get_file_name,
    get_file_name_with_ext,
    silent_remove,
)

Image.MAX_IMAGE_PIXELS = None
EXT_TO_CONVERT = [".heic", ".avif", ".heif", ".jfif"]


def validate_image(path: str) -> tuple:
    """Validate image by mimetypes, ext, and remove alpha channel."""

    try:
        ext = get_file_ext(path)
        if ext.lower() == ".nrrd":
            return path
        if ext.lower() in EXT_TO_CONVERT:
            path = convert_to_jpg(path)
            return path
        if ext.lower() != ".mpo":
            name = get_file_name_with_ext(path)
            new_name = validate_mimetypes(name, path)
            if new_name != name:
                img = read(path, remove_alpha_channel=False)
                silent_remove(path)
                path = Path(path).with_name(new_name).as_posix()
                write(path, img, remove_alpha_channel=False)
        return path
    except Exception as e:
        logger.warning(f"Skip image: {repr(e)}", extra={"file_path": path})


def validate_mimetypes(name: str, path: str) -> list:
    """Validate mimetypes for images."""

    mimetypes.add_type("image/webp", ".webp")  # to extend types_map
    mimetypes.add_type("image/jpeg", ".jfif")  # to extend types_map

    with open(path, "rb") as f:
        mimetype = magic.from_buffer(f.read(), mime=True)
    file_ext = get_file_ext(path).lower()
    if file_ext in mimetypes.guess_all_extensions(mimetype):
        return name

    new_img_ext = mimetypes.guess_extension(mimetype)
    if new_img_ext == ".bin" or new_img_ext is None:
        new_img_ext = ".jpeg"
    new_img_name = f"{get_file_name(name)}{new_img_ext}"
    logger.info(f"Image {name} with mimetype {mimetype} will be converted to {new_img_ext}")

    return new_img_name


def convert_to_jpg(path) -> tuple:
    """Convert image to jpg."""

    # * do not remove folllowing imports, it is used to register avif/heic formats
    import pillow_avif  # pylint: disable=import-error
    from pillow_heif import register_heif_opener  # pylint: disable=import-error

    register_heif_opener()

    new_path = Path(path).with_suffix(".jpeg").as_posix()
    with Image.open(path) as image:
        image.convert("RGB").save(new_path)
    silent_remove(path)
    return new_path


def read_tiff_image(path: str) -> Union[np.ndarray, None]:
    """
    Read tiff image.
    Method will transpose image if it has shape (C, H, W) to (H, W, C).
    """

    import tifffile

    logger.debug(f"Found tiff file: {path}.")
    try:
        image = tifffile.imread(path)
    except Exception as e:
        logger.warning(
            f"tifffile failed to read TIFF, trying Pillow fallback: {repr(e)}",
            extra={"file_path": path},
        )
        image = _read_tiff_image_fallback(path)
    name = get_file_name_with_ext(path)
    if image is not None:
        tiff_shape = image.shape
        if image.ndim == 3:
            if tiff_shape[0] < tiff_shape[1] and tiff_shape[0] < tiff_shape[2]:
                image = image.transpose(1, 2, 0)
                logger.warning(f"{name}: transposed shape from {tiff_shape} to {image.shape}")

    return image


def _read_tiff_image_fallback(path: str) -> Union[np.ndarray, None]:
    """
    Fallback method to read tiff image using Pillow.
    """
    from PIL import ImageSequence

    try:
        with Image.open(path) as pil_img:
            frames = [np.asarray(frame) for frame in ImageSequence.Iterator(pil_img)]
        if not frames:
            return None
        if len(frames) == 1:
            return frames[0]

        if all(frame.shape == frames[0].shape for frame in frames):
            return np.stack(frames, axis=0)

        logger.warning(
            "TIFF has multiple pages with different shapes; using the first page only.",
            extra={"file_path": path},
        )
        return frames[0]
    except Exception as e:
        logger.warning(
            f"Pillow failed to read TIFF: {repr(e)}",
            extra={"file_path": path},
        )
        return None


def validate_image_bounds(labels: List[Label], img_rect: Rectangle) -> List[Label]:
    """
    Check if labels are localed inside the image canvas, print a warning and skip them if not.
    """
    new_labels = []
    for label in labels:
        if isinstance(label.geometry, OrientedBBox):
            new_labels.append(label)
        elif img_rect.contains(label.geometry.to_bbox()):
            new_labels.append(label)
    if new_labels != labels:
        logger.warning(
            f"{len(labels) - len(new_labels)} annotation objects are out of image bounds. Skipping..."
        )
    return new_labels



import os
import uuid
from typing import Dict, List, Optional

import supervisely.convert.pointcloud_episodes.nuscenes_conv.nuscenes_helper as helpers
import supervisely.io.fs as fs
from supervisely import KeyIdMap, PointcloudAnnotation, PointcloudObject
from supervisely._utils import is_development
from supervisely.annotation.obj_class import ObjClass
from supervisely.annotation.tag_meta import TagMeta, TagValueType
from supervisely.api.api import Api, ApiField
from supervisely.api.dataset_api import DatasetInfo
from supervisely.convert.base_converter import AvailablePointcloudConverters
from supervisely.convert.pointcloud.pointcloud_converter import PointcloudConverter
from supervisely.convert.pointcloud_episodes.nuscenes_conv.nuscenes_converter import (
    NuscenesEpisodesConverter,
)
from supervisely.geometry.cuboid_3d import Cuboid3d
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
from supervisely.pointcloud_annotation.pointcloud_object_collection import (
    PointcloudObjectCollection,
)
from supervisely.pointcloud_annotation.pointcloud_tag import PointcloudTag
from supervisely.pointcloud_annotation.pointcloud_tag_collection import (
    PointcloudTagCollection,
)
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger


class NuscenesConverter(NuscenesEpisodesConverter, PointcloudConverter):
    """Converter for NuScenes pointcloud format."""

    def to_supervisely(
        self,
        scene_sample: helpers.Sample,
        meta: ProjectMeta,
        renamed_classes: dict = {},
        renamed_tags: dict = {},
    ) -> PointcloudAnnotation:
        bevbox_objs = [obj.convert_nuscenes_to_BEVBox3D() for obj in scene_sample.anns]
        geoms = [obj.to_supervisely() for obj in scene_sample.anns]
        attrs = [obj.attributes for obj in scene_sample.anns]

        figures = []
        objs = []
        for label, geom, attributes in zip(bevbox_objs, geoms, attrs):
            class_name = renamed_classes.get(label.label_class, label.label_class)
            tag_col = None
            if len(attributes) > 0 and all([tag_name is not None for tag_name in attributes]):
                tag_meta_names = [renamed_tags.get(name, name) for name in attributes]
                tag_metas = [meta.get_tag_meta(tag_meta_name) for tag_meta_name in tag_meta_names]
                tag_col = PointcloudTagCollection([PointcloudTag(meta, None) for meta in tag_metas])
            pcobj = PointcloudObject(meta.get_obj_class(class_name), tag_col)
            figures.append(PointcloudFigure(pcobj, geom))
            objs.append(pcobj)
        return PointcloudAnnotation(PointcloudObjectCollection(objs), figures)

    def upload_dataset(self, api: Api, dataset_id: int, batch_size: int = 1, log_progress=True):
        from nuscenes.nuscenes import NuScenes  # pylint: disable=import-error

        nuscenes: NuScenes = self._nuscenes

        key_id_map = KeyIdMap()

        tag_metas = [TagMeta(attr["name"], TagValueType.NONE) for attr in nuscenes.attribute]
        obj_classes = []
        classes_token_map = {}
        for category in nuscenes.category:
            color = nuscenes.colormap[category["name"]]
            description = helpers.trim_description(category["description"])
            obj_classes.append(ObjClass(category["name"], Cuboid3d, color, description=description))
            classes_token_map[category["token"]] = category["name"]
        self._custom_data["classes_token_map"] = classes_token_map

        self._meta = ProjectMeta(obj_classes, tag_metas)
        meta, renamed_classes, renamed_tags = self.merge_metas_with_conflicts(api, dataset_id)

        dataset_info = api.dataset.get_info_by_id(dataset_id)
        scene_name_to_dataset: Dict[str, DatasetInfo] = {}

        scene_names = [scene["name"] for scene in nuscenes.scene]
        scene_cnt = len(scene_names)
        total_sample_cnt = sum([scene["nbr_samples"] for scene in nuscenes.scene])

        multiple_scenes = len(scene_names) > 1
        if multiple_scenes:
            logger.info(f"Found {scene_cnt} scenes ({total_sample_cnt} samples) in the input data.")
            # * Create a nested dataset for each scene
            for name in scene_names:
                ds = api.dataset.create(
                    dataset_info.project_id,
                    name,
                    change_name_if_conflict=True,
                    parent_id=dataset_id,
                )
                scene_name_to_dataset[name] = ds
        else:
            scene_name_to_dataset[scene_names[0]] = dataset_info

        if log_progress:
            progress, progress_cb = self.get_progress(total_sample_cnt, "Converting pointclouds...")
        else:
            progress_cb = None

        self._custom_data["frame_token_map"] = {}
        for scene in nuscenes.scene:
            current_dataset_id = scene_name_to_dataset[scene["name"]].id

            log = nuscenes.get("log", scene["log_token"])
            sample_token = scene["first_sample_token"]

            # * Extract scene's samples
            scene_samples: List[helpers.Sample] = []
            frame_token_map = {}
            for i in range(scene["nbr_samples"]):
                sample = nuscenes.get("sample", sample_token)
                frame_token_map[sample["token"]] = i
                lidar_path, boxes, _ = nuscenes.get_sample_data(sample["data"]["LIDAR_TOP"])
                if not os.path.exists(lidar_path):
                    logger.warning(f'Scene "{scene["name"]}" has no LIDAR data.')
                    continue

                timestamp = sample["timestamp"]
                anns = []
                for box, name, inst_token in helpers.Sample.generate_boxes(nuscenes, boxes):
                    current_instance_token = inst_token["token"]
                    parent_token = inst_token["prev"]

                    # get category, attributes and visibility
                    ann = nuscenes.get("sample_annotation", current_instance_token)
                    category = ann["category_name"]
                    attributes = [
                        nuscenes.get("attribute", attr)["name"] for attr in ann["attribute_tokens"]
                    ]
                    visibility = nuscenes.get("visibility", ann["visibility_token"])["level"]

                    ann_uuid = uuid.UUID(ann["token"])
                    ann = helpers.AnnotationObject(
                        name=name,
                        bbox=box,
                        token=ann_uuid,
                        instance_token=current_instance_token,
                        parent_token=parent_token,
                        category=category,
                        attributes=attributes,
                        visibility=visibility,
                    )
                    anns.append(ann)

                # get camera data
                sample_data = nuscenes.get("sample_data", sample["data"]["LIDAR_TOP"])
                cal_sensor = nuscenes.get(
                    "calibrated_sensor", sample_data["calibrated_sensor_token"]
                )
                ego_pose = nuscenes.get("ego_pose", sample_data["ego_pose_token"])

                camera_data = [
                    helpers.CamData(nuscenes, sensor, token, cal_sensor, ego_pose)
                    for sensor, token in sample["data"].items()
                    if sensor.startswith("CAM")
                ]
                scene_samples.append(helpers.Sample(timestamp, lidar_path, anns, camera_data))
                sample_token = sample["next"]
            self._custom_data["frame_token_map"][current_dataset_id] = frame_token_map

            # * Convert and upload pointclouds w/ annotations
            for idx, sample in enumerate(scene_samples):
                pcd_ann = self.to_supervisely(sample, meta, renamed_classes, renamed_tags)

                pcd_path = sample.convert_lidar_to_supervisely()
                pcd_name = fs.get_file_name_with_ext(pcd_path)
                pcd_meta = {
                    "frame": idx,
                    "vehicle": log["vehicle"],
                    "date": log["date_captured"],
                    "location": log["location"],
                    "description": scene["description"],
                }
                info = api.pointcloud.upload_path(current_dataset_id, pcd_name, pcd_path, pcd_meta)
                fs.silent_remove(pcd_path)

                pcd_id = info.id
                # * Upload pointcloud annotation
                try:
                    api.pointcloud.annotation.append(pcd_id, pcd_ann, key_id_map)
                except Exception as e:
                    error_msg = getattr(getattr(e, "response", e), "text", str(e))
                    logger.warning(
                        f"Failed to upload annotation for scene: {scene['name']}. Message: {error_msg}"
                    )

                # * Upload related images
                image_jsons = []
                camera_names = []
                for img_path, rimage_info in [
                    data.get_info(sample.timestamp) for data in sample.cam_data
                ]:
                    img = api.pointcloud.upload_related_image(img_path)
                    image_jsons.append(
                        {
                            ApiField.ENTITY_ID: pcd_id,
                            ApiField.NAME: rimage_info[ApiField.NAME],
                            ApiField.HASH: img,
                            ApiField.META: rimage_info[ApiField.META],
                        }
                    )
                    camera_names.append(rimage_info[ApiField.META]["deviceId"])
                if len(image_jsons) > 0:
                    api.pointcloud.add_related_images(image_jsons, camera_names)

                if log_progress:
                    progress_cb(1)

            logger.info(f"Dataset ID:{current_dataset_id} has been successfully uploaded.")

        key_id_map = key_id_map.to_dict()
        key_id_map.pop("tags")
        key_id_map.pop("videos")
        self._custom_data["key_id_map"] = key_id_map

        project_id = dataset_info.project_id
        current_custom_data = api.project.get_custom_data(project_id)
        current_custom_data.update(self._custom_data)
        api.project.update_custom_data(project_id, current_custom_data)

        if log_progress:
            if is_development():
                progress.close()



# coding: utf-8
from supervisely.collection.str_enum import StrEnum


# ['admin',  'developer', 'manager', 'reviewer', 'annotator', 'viewer']
class UserRoleName(StrEnum):
    """
    Enumerates supported Supervisely user roles (e.g. admin, manager, annotator).

    You can learn more about the roles in the `User Roles documentation <https://docs.supervisely.com/collaboration/members>`_.
    """
    ADMIN = 'admin'
    """Has full access in the team. Admin can invite new team members and remove entities created by the other team members."""
    DEVELOPER = 'developer'
    """Similar to the admin, but can only remove entities created by themself and cannot invite new members to the team."""
    MANAGER = 'manager'
    """Has no access to things like Neural Networks, but can view and modify Projects & Labeling Jobs."""
    REVIEWER = 'reviewer'
    """Same as Annotator, but can also create new labeling jobs."""
    ANNOTATOR = 'annotator'
    """Has access only to a single page, Labeling Jobs."""
    VIEWER = 'viewer'
    """Can only view items in team."""


# coding: utf-8

# docs
from typing import Callable, Dict, List, Optional, Union

from tqdm import tqdm

from supervisely._utils import batched
from supervisely.api.module_api import ApiField, ModuleApi
from supervisely.video_annotation.key_id_map import KeyIdMap


class EntityAnnotationAPI(ModuleApi):
    """Base API module for downloading/appending annotations for an entity type (image, video, pointcloud, etc.)."""

    _method_download_bulk = None
    """"""
    _entity_ids_str = None
    """"""

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`
        """
        super().__init__(api)
        self._api = api

    def download(self, entity_id):
        """"""
        raise NotImplementedError()

    def _download(self, dataset_id, entity_id):
        """
        :param dataset_id: int
        :param entity_id: int
        :returns: list of content(annotation with given id from dataset with given id), received after execution post request
        """
        return self.download_bulk(dataset_id, [entity_id])[0]

    def download_bulk(self, dataset_id: int, entity_ids: List[int]) -> Dict:
        """
        Download a list of content (annotations with given ids from dataset with given id).

        :param dataset_id: int
        :param entity_ids: list of integers
        :returns: list of content(annotations with given ids from dataset with given id), received after execution post request
        """
        response = self._api.post(
            self._method_download_bulk,
            {ApiField.DATASET_ID: dataset_id, self._entity_ids_str: entity_ids},
        )
        return response.json()

    def _append(
        self,
        tag_api,
        object_api,
        figure_api,
        project_id,
        dataset_id,
        entity_id,
        tags,
        objects,
        figures,
        key_id_map: KeyIdMap = None,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ):
        """"""
        if key_id_map is None:
            # create for internal purposes (to link figures and tags to objects)
            key_id_map = KeyIdMap()

        tag_api.append_to_entity(entity_id, project_id, tags, key_id_map=key_id_map)
        object_api.append_bulk(entity_id, objects, key_id_map)
        for fig_batch in batched(figures, batch_size=1000):
            figure_api.append_bulk(entity_id, fig_batch, key_id_map)
            if progress_cb is not None:
                if hasattr(progress_cb, "update") and callable(getattr(progress_cb, "update")):
                    progress_cb.update(len(fig_batch))
                else:
                    progress_cb(len(fig_batch))

    def append(self, entity_id, ann, key_id_map: KeyIdMap = None):
        """"""
        raise NotImplementedError()

    # def append(self, video_id, ann, key_id_map: KeyIdMap = None):
    #     if key_id_map is None:
    #         # create for internal purposes (to link figures and tags to objects)
    #         key_id_map = KeyIdMap()
    #
    #     info = self._api.video.get_info_by_id(video_id)
    #
    #     self._api.tag.append_to_video(video_id, ann.tags, key_id_map=key_id_map)
    #     self._api.object.append_bulk(video_id, info.project_id, info.dataset_id, ann.objects, key_id_map)
    #     self._api.video.figure.append_bulk(video_id, ann.frames, key_id_map)



from supervisely.nn.benchmark.base_visualizer import BaseVisMetrics
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class ClasswiseErrorAnalysis(BaseVisMetrics):
    """Per-class error analysis chart for segmentation model comparison."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.clickable = True

    @property
    def md(self) -> MarkdownWidget:
        return MarkdownWidget(
            "classwise_error_analysis",
            "Classwise Segmentation Error Analysis",
            text=self.vis_texts.markdown_eou_per_class,
        )

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget("classwise_error_analysis", self.get_figure())

    def get_figure(self):
        import numpy as np
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()

        # Цветовая палитра для метрик
        color_palette = ["#8ACAA1", "#FFE4B5", "#F7ADAA", "#dd3f3f"]
        metrics = ["IoU", "E_extent_oU", "E_boundary_oU", "E_segment_oU"]

        group_width = 0.7

        for model_idx, eval_result in enumerate(self.eval_results):
            bar_data, labels = eval_result.mp.classwise_segm_error_data
            model_name = eval_result.name

            for metric_idx, metric_name in enumerate(metrics):
                # hover_customdata = [f"metric: {metric_name} for class '{l}' ({model_name})" for l in labels]
                hover_customdata = [
                    f"class: {l}<br>model: {model_name}<br>{metric_name}" for l in labels
                ]
                fig.add_trace(
                    go.Bar(
                        name=metric_name,
                        x=np.arange(len(labels)) + model_idx * group_width * 0.3,
                        y=bar_data[metric_name],
                        customdata=hover_customdata,
                        hovertemplate="%{customdata}: %{y:.2f}<extra></extra>",
                        marker=dict(color=color_palette[metric_idx]),
                        width=group_width / len(metrics),
                        offsetgroup=model_idx,
                        base=bar_data[metrics[:metric_idx]].sum(axis=1) if metric_idx > 0 else None,
                    )
                )

        fig.update_layout(
            showlegend=False,
            barmode="stack",
            xaxis=dict(
                title="Classes",
                tickvals=np.arange(len(labels)) + (len(self.eval_results) - 1 - group_width) / 4,
                ticktext=labels,
            ),
            width=800 if len(labels) < 10 else 1000,
        )

        return fig



import os
from collections import defaultdict
from typing import Dict, Generator, Union

import cv2
import numpy as np
import tqdm

from supervisely import ProjectMeta, generate_free_name, is_development, logger
from supervisely._utils import batched
from supervisely.annotation.annotation import Annotation
from supervisely.api.api import Api
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.high_color import high_color_helper as helpers
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.io.env import task_id as get_task_id
from supervisely.io.env import team_id as get_team_id
from supervisely.io.fs import list_files_recursively
from supervisely.io.json import load_json_file
from supervisely.project.project_settings import LabelingInterface
from supervisely.team_files import RECOMMENDED_IMPORT_BACKUP_PATH


class HighColorDepthImageConverter(ImageConverter):
    """Converter for high color-depth images (e.g. EXR/HDR) with optional backup and NRRD conversion."""

    # allowed_exts = [".png", ".tiff", ".tif", ".bmp", ".exr", ".hdr"]
    allowed_exts = [".exr", ".hdr"]

    class Item(ImageConverter.Item):
        """Image item extended with a reference to the backed-up original file path."""

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._original_path = None

        @property
        def original_path(self) -> Union[str, None]:
            return self._original_path

        @original_path.setter
        def original_path(self, original_path: str) -> None:
            self._original_path = original_path

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._team_id = None

    def __str__(self):
        return AvailableImageConverters.HIGH_COLOR_DEPTH

    def validate_labeling_interface(self) -> bool:
        """Only multi_view labeling interface can be used to group images on single screen."""
        return self._labeling_interface in [
            LabelingInterface.DEFAULT,
            LabelingInterface.IMAGES_WITH_16_COLOR,
        ]

    def validate_format(self) -> bool:
        logger.debug(f"Validating format: {self.__str__()}")

        for image_path in self._find_high_color_depth_image_generator(self._input_data):
            item = self.Item(image_path)
            self._items.append(item)

        if self.items_count == 0:
            logger.debug(f"Not found any images with high color depth in '{self._input_data}'")
            return False

        return True

    def _find_high_color_depth_image_generator(self, input_data: str) -> Generator[str, None, None]:
        for image_path in list_files_recursively(input_data, valid_extensions=self.allowed_exts):
            if self._is_high_color_depth_image(image_path):
                yield image_path

    def _is_high_color_depth_image(self, image_path: str) -> bool:
        image = helpers.read_high_color_images(image_path)
        if image is None:
            return False
        return image.dtype in self._supported_depths

    @property
    def _supported_depths(self):
        return [np.uint16, np.int16, np.float16, np.float32, np.float64, np.uint32]

    def to_supervisely(self, *args, **kwargs):
        return None

    def upload_dataset(
        self,
        api: Api,
        dataset_id: int,
        batch_size: int = 50,
        log_progress=True,
    ) -> None:
        """Upload converted data to Supervisely"""
        team_id = get_team_id()
        backup_dir = os.path.join(RECOMMENDED_IMPORT_BACKUP_PATH, str(get_task_id()))

        # progress, progress_cb = None, None
        # if log_progress:
        #     progress, progress_cb = self.get_progress(
        #         len(self.get_items()),
        #         "Backing up original files and converting images to nrrd...",
        #     )

        # Back up original files to avoid data loss and convert images to nrrd format
        with tqdm.tqdm(
            total=len(self.get_items()),
            desc="Backing up original files and converting images to nrrd...",
        ) as pbar:
            for batch_items in batched(self.get_items(), batch_size=batch_size):

                local_paths, remote_paths = [], []

                for item in batch_items:
                    local_paths.append(item.path)

                    remote_path = os.path.join(backup_dir, item.name)
                    remote_paths.append(remote_path)

                    item.original_path = remote_path
                    image = helpers.read_high_color_images(item.path)

                    orig_item_name = item.name
                    nrrd_path = item.path + ".nrrd"
                    nrrd_path = helpers.save_nrrd(image, nrrd_path)
                    item.path = nrrd_path

                    item_meta = {}
                    # Add original file path to image meta
                    item_meta["original_file_path"] = os.path.join(backup_dir, orig_item_name)
                    if item.meta:
                        item_meta.update(item.meta)
                    item.set_meta_data(item_meta)

                api.file.upload_bulk(team_id, local_paths, remote_paths)
                pbar.update(len(batch_items))

        if log_progress and is_development():
            pbar.close()

        super().upload_dataset(api, dataset_id, batch_size, log_progress, self.get_items())



from datetime import datetime
from typing import Any, Callable, Dict, List, Literal, Optional, Union

from supervisely.app import StateJson
from supervisely.app.widgets import Widget


class DatePicker(Widget):
    """Date picker input."""

    class Routes:
        """Route name constants for this widget."""
        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        value: Optional[Union[int, str, list, tuple]] = None,
        placeholder: Optional[str] = "Select date",
        picker_type: Optional[
            Literal["year", "month", "date", "datetime", "week", "datetimerange", "daterange"]
        ] = "date",
        size: Optional[Literal["large", "small", "mini"]] = None,
        readonly: Optional[bool] = False,
        disabled: Optional[bool] = False,
        editable: Optional[bool] = False,
        clearable: Optional[bool] = True,
        format: Optional[Literal["yyyy", "MM", "dd", "HH", "mm", "ss"]] = "yyyy-MM-dd",
        first_day_of_week: Optional[int] = 1,
        widget_id: Optional[str] = None,
    ):
        """
        :param value: Initial value (datetime, str, or timestamp).
        :type value: Optional[Union[int, str, list, tuple]]
        :param placeholder: Placeholder text.
        :type placeholder: Optional[str]
        :param picker_type: year, month, date, datetime, week, datetimerange, daterange.
        :type picker_type: Optional[Literal["year", "month", "date", "datetime", "week", "datetimerange", "daterange"]]
        :param size: large, small, or mini.
        :type size: Optional[Literal["large", "small", "mini"]]
        :param readonly: Read-only mode.
        :type readonly: Optional[bool]
        :param disabled: Disabled state.
        :type disabled: Optional[bool]
        :param editable: Allow manual edit.
        :type editable: Optional[bool]
        :param clearable: Allow clearing.
        :type clearable: Optional[bool]
        :param format: Date format string.
        :type format: Optional[Literal["yyyy", "MM", "dd", "HH", "mm", "ss"]]
        :param first_day_of_week: 0–6 (0=Sunday).
        :type first_day_of_week: Optional[int]
        :param widget_id: Unique widget identifier.
        :type widget_id: Optional[str]

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import DatePicker
                dp = DatePicker(value=datetime.now(), picker_type="date", placeholder="Select date")
        """
        self._value = value
        self._readonly = readonly
        self._picker_type = picker_type
        self._size = size
        self._disabled = disabled
        self._placeholder = placeholder
        self._editable = editable
        self._clearable = clearable
        self._format = format
        self._first_day_of_week = first_day_of_week
        self._changes_handled = False

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, Union[str, bool, int, Dict[str, int]]]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - type: string representation of the widget type (year, month, date, datetime, week, datetimerange, daterange)
            - size: size of the input box, one of: large, small, mini
            - readonly: if True, date picker will be readonly
            - disabled: if True, date picker will be disabled
            - editable: if True, date picker will be editable
            - clearable: if True, date picker will be clearable
            - placeholder: placeholder text
            - format: date format, one of: yyyy, MM, dd, HH, mm, ss
            - options: dictionary with options for date picker
                - firstDayOfWeek: first day of week

        :returns: dictionary with widget data
        :rtype: Dict[str, Union[str, bool, int, Dict[str, int]]]
        """

        return {
            "type": self._picker_type,
            "size": self._size,
            "readonly": self._readonly,
            "disabled": self._disabled,
            "editable": self._editable,
            "clearable": self._clearable,
            "placeholder": self._placeholder,
            "format": self._format,
            "options": {"firstDayOfWeek": self._first_day_of_week},
        }

    def get_json_state(self) -> Dict[str, Union[str, List[int]]]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - value: current value

        :returns: dictionary with widget state
        :rtype: Dict[str, Union[str, List[int]]]
        """
        return {"value": self._value}

    def clear_value(self) -> None:
        """Clears current value of the widget."""
        self._value = None
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def get_value(self) -> Union[int, str, datetime, list, tuple, None]:
        """Returns current value.

        :returns: current value
        :rtype: Union[int, str, datetime, list, tuple, None]
        """
        if "value" not in StateJson()[self.widget_id].keys():
            return None
        value = StateJson()[self.widget_id]["value"]
        if self._picker_type in ["datetimerange", "daterange"] and any(
            [bool(date) is False for date in value]
        ):
            return None
        elif self._picker_type not in ["datetimerange", "daterange"] and value == "":
            return None
        return value

    def set_value(self, value: Union[int, str, datetime, list, tuple]) -> None:
        """Sets current value.

        :param value: current value
        :type value: Union[int, str, datetime, list, tuple]
        """
        if self._picker_type in ["year", "month", "date", "datetime", "week"]:
            if type(value) not in [int, str, datetime]:
                raise ValueError(
                    f'Date picker type "{self._picker_type}" does not support value "{value}" of type: "{str(type(value))}". Value type has to be one of: ["int", "str", "datetime].'
                )
            if isinstance(value, datetime):
                value = str(value)

        if self._picker_type in ["datetimerange", "daterange"]:
            if type(value) not in [list, tuple]:
                raise ValueError(
                    f'Date picker type "{self._picker_type}" does not support value "{value}" of type: "{str(type(value))}". Value type has to be one of: ["list", "tuple"].'
                )
            if len(value) != 2:
                raise ValueError(f"Value length has to be equal 2: {len(value)} != 2")
            value = [str(val) if isinstance(val, datetime) else val for val in value]

        self._value = value
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def value_changed(
        self, func: Callable[[Union[int, str, datetime, list, tuple]], Any]
    ) -> Callable[[], None]:
        """Decorator fot the function that will be called when value is changed.

        :param func: function that will be called when value is changed
        :type func: Callable[[Union[int, str, datetime, list, tuple]], Any]
        :returns: decorated function
        :rtype: Callable[[], None]
        """
        route_path = self.get_route_path(DatePicker.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _value_changed():
            res = self.get_value()
            func(res)

        return _value_changed
